from owslib.etree import etree
from owslib.util import Authentication, openURL

from urllib.parse import urlencode, parse_qsl


class WFSCapabilitiesReader(object):
    """Read and parse capabilities document into a lxml.etree infoset
    """

    def __init__(self, version="1.0", username=None, password=None, headers=None, auth=None):
        """Initialize"""
        self.headers = headers
        if auth:
            if username:
                auth.username = username
            if password:
                auth.password = password
        self.auth = auth or Authentication(username, password)
        self.version = version
        self._infoset = None

    def capabilities_url(self, service_url):
        """Return a capabilities url
        """
        qs = []
        if service_url.find("?") != -1:
            qs = parse_qsl(service_url.split("?")[1])

        params = [x[0] for x in qs]

        if "service" not in params:
            qs.append(("service", "WFS"))
        if "request" not in params:
            qs.append(("request", "GetCapabilities"))
        if "version" not in params:
            qs.append(("version", self.version))

        urlqs = urlencode(tuple(qs))
        return service_url.split("?")[0] + "?" + urlqs

    def read(self, url, timeout=30):
        """Get and parse a WFS capabilities document, returning an
        instance of WFSCapabilitiesInfoset

        Parameters
        ----------
        url : string
            The URL to the WFS capabilities document.
        timeout : number
            A timeout value (in seconds) for the request.
        """
        request = self.capabilities_url(url)
        u = openURL(request, timeout=timeout, headers=self.headers, auth=self.auth)
        return etree.fromstring(u.read())

    def readString(self, st):
        """Parse a WFS capabilities document, returning an
        instance of WFSCapabilitiesInfoset

        string should be an XML capabilities document
        """
        if not isinstance(st, str) and not isinstance(st, bytes):
            raise ValueError(
                "String must be of type string or bytes, not %s" % type(st)
            )
        return etree.fromstring(st)


class AbstractContentMetadata(object):
    def __init__(self, headers=None, auth=None):
        self.auth = auth or Authentication()
        self.headers = headers

    def get_metadata(self):
        return [
            m["metadata"]
            for m in self.metadataUrls
            if m.get("metadata", None) is not None
        ]



# =============================================================================
# OWSLib. Copyright (C) 2005 Sean C. Gillies
#
# Contact email: sgillies@frii.com
#
# $Id: wfs.py 503 2006-02-01 17:09:12Z dokai $
# =============================================================================

from owslib import util

from io import BytesIO
from urllib.parse import urlencode
from owslib.util import (
    testXMLValue,
    extract_xml_list,
    ServiceException,
    xmltag_split,
    Authentication,
    openURL,
    log,
)
from owslib.etree import etree
from owslib.fgdc import Metadata
from owslib.iso import MD_Metadata
from owslib.crs import Crs
from owslib.namespaces import Namespaces
from owslib.feature.schema import get_schema
from owslib.feature.common import (
    WFSCapabilitiesReader,
    AbstractContentMetadata,
)

import pyproj

n = Namespaces()
WFS_NAMESPACE = n.get_namespace("wfs")
OGC_NAMESPACE = n.get_namespace("ogc")


# TODO: use nspath in util.py
def nspath(path, ns=WFS_NAMESPACE):
    """
    Prefix the given path with the given namespace identifier.

    Parameters
    ----------
    path : string
        ElementTree API Compatible path expression

    ns : string
        The XML namespace. Defaults to WFS namespace.
    """
    components = []
    for component in path.split("/"):
        if component != "*":
            component = "{%s}%s" % (ns, component)
        components.append(component)
    return "/".join(components)


class WebFeatureService_1_0_0(object):
    """Abstraction for OGC Web Feature Service (WFS).

    Implements IWebFeatureService.
    """

    def __new__(
        self,
        url,
        version,
        xml,
        parse_remote_metadata=False,
        timeout=30,
        headers=None,
        username=None,
        password=None,
        auth=None,
    ):
        """ overridden __new__ method

        @type url: string
        @param url: url of WFS capabilities document
        @type xml: string
        @param xml: elementtree object
        @type parse_remote_metadata: boolean
        @param parse_remote_metadata: whether to fully process MetadataURL elements
        @param headers: HTTP headers to send with requests
        @param timeout: time (in seconds) after which requests should timeout
        @param username: service authentication username
        @param password: service authentication password
        @param auth: instance of owslib.util.Authentication
        @return: initialized WebFeatureService_1_0_0 object
        """
        obj = object.__new__(self)
        obj.__init__(
            url,
            version,
            xml,
            parse_remote_metadata,
            timeout,
            headers=headers,
            username=username,
            password=password,
            auth=auth,
        )
        return obj

    def __getitem__(self, name):
        """ check contents dictionary to allow dict like access to service layers"""
        if name in list(self.__getattribute__("contents").keys()):
            return self.__getattribute__("contents")[name]
        else:
            raise KeyError("No content named %s" % name)

    def __init__(
        self,
        url,
        version,
        xml=None,
        parse_remote_metadata=False,
        timeout=30,
        headers=None,
        username=None,
        password=None,
        auth=None,
    ):
        """Initialize."""
        if auth:
            if username:
                auth.username = username
            if password:
                auth.password = password
        self.url = url
        self.version = version
        self.timeout = timeout
        self.headers = headers
        self.auth = auth or Authentication(username, password)
        self._capabilities = None
        reader = WFSCapabilitiesReader(self.version, headers=self.headers, auth=self.auth)
        if xml:
            self._capabilities = reader.readString(xml)
        else:
            self._capabilities = reader.read(self.url)
        self._buildMetadata(parse_remote_metadata)

    def _buildMetadata(self, parse_remote_metadata=False):
        """set up capabilities metadata objects: """

        self.updateSequence = self._capabilities.attrib.get("updateSequence")

        # serviceIdentification metadata
        serviceelem = self._capabilities.find(nspath("Service"))
        self.identification = ServiceIdentification(serviceelem, self.version)

        # serviceProvider metadata
        self.provider = ServiceProvider(serviceelem)

        # serviceOperations metadata
        self.operations = []
        for elem in self._capabilities.find(nspath("Capability/Request"))[:]:
            self.operations.append(OperationMetadata(elem))

        # serviceContents metadata: our assumption is that services use a top-level
        # layer as a metadata organizer, nothing more.

        self.contents = {}
        featuretypelist = self._capabilities.find(nspath("FeatureTypeList"))
        features = self._capabilities.findall(nspath("FeatureTypeList/FeatureType"))
        for feature in features:
            cm = ContentMetadata(
                feature, featuretypelist, parse_remote_metadata, auth=self.auth
            )
            self.contents[cm.id] = cm

        # exceptions
        self.exceptions = [
            f.text for f in self._capabilities.findall("Capability/Exception/Format")
        ]

    def getcapabilities(self):
        """Request and return capabilities document from the WFS as a
        file-like object.
        NOTE: this is effectively redundant now"""
        reader = WFSCapabilitiesReader(self.version, auth=self.auth)
        return openURL(
            reader.capabilities_url(self.url), timeout=self.timeout,
            headers=self.headers, auth=self.auth
        )

    def items(self):
        """supports dict-like items() access"""
        items = []
        for item in self.contents:
            items.append((item, self.contents[item]))
        return items

    def getfeature(
        self,
        typename=None,
        filter=None,
        bbox=None,
        featureid=None,
        featureversion=None,
        propertyname="*",
        maxfeatures=None,
        srsname=None,
        outputFormat=None,
        method="{http://www.opengis.net/wfs}Get",
        startindex=None,
    ):
        """Request and return feature data as a file-like object.

        Parameters
        ----------
        typename : list
            List of typenames (string)
        filter : string
            XML-encoded OGC filter expression.
        bbox : tuple
            (left, bottom, right, top) in the feature type's coordinates.
        featureid : list
            List of unique feature ids (string)
        featureversion : string
            Default is most recent feature version.
        propertyname : list
            List of feature property names. '*' matches all.
        maxfeatures : int
            Maximum number of features to be returned.
        method : string
            Qualified name of the HTTP DCP method to use.
        srsname: string
            EPSG code to request the data in
        outputFormat: string (optional)
            Requested response format of the request.
        startindex: int (optional)
            Start position to return feature set (paging in combination with maxfeatures)


        There are 3 different modes of use

        1) typename and bbox (simple spatial query)
        2) typename and filter (more expressive)
        3) featureid (direct access to known features)
        """
        try:
            base_url = next(
                (
                    m.get("url")
                    for m in self.getOperationByName("GetFeature").methods
                    if m.get("type").lower() == method.lower()
                )
            )
        except StopIteration:
            base_url = self.url
        request = {"service": "WFS", "version": self.version, "request": "GetFeature"}

        # check featureid
        if featureid:
            request["featureid"] = ",".join(featureid)
        elif bbox and typename:
            request["bbox"] = ",".join([repr(x) for x in bbox])
        elif filter and typename:
            request["filter"] = str(filter)

        if srsname:
            request["srsname"] = str(srsname)

        assert len(typename) > 0
        request["typename"] = ",".join(typename)

        if propertyname is not None:
            if not isinstance(propertyname, list):
                propertyname = [propertyname]
            request["propertyname"] = ",".join(propertyname)

        if featureversion:
            request["featureversion"] = str(featureversion)
        if maxfeatures:
            request["maxfeatures"] = str(maxfeatures)
        if startindex:
            request["startindex"] = str(startindex)

        if outputFormat is not None:
            request["outputFormat"] = outputFormat

        data = urlencode(request)
        log.debug("Making request: %s?%s" % (base_url, data))
        u = openURL(base_url, data, method, timeout=self.timeout,
                    headers=self.headers, auth=self.auth)

        # check for service exceptions, rewrap, and return
        # We're going to assume that anything with a content-length > 32k
        # is data. We'll check anything smaller.

        if "Content-Length" in u.info():
            length = int(u.info()["Content-Length"])
            have_read = False
        else:
            data = u.read()
            have_read = True
            length = len(data)

        if length < 32000:
            if not have_read:
                data = u.read()

            try:
                tree = etree.fromstring(data)
            except BaseException:
                # Not XML
                return BytesIO(data)
            else:
                if tree.tag == "{%s}ServiceExceptionReport" % OGC_NAMESPACE:
                    se = tree.find(nspath("ServiceException", OGC_NAMESPACE))
                    raise ServiceException(str(se.text).strip())
                else:
                    return BytesIO(data)
        else:
            if have_read:
                return BytesIO(data)
            return u

    def getOperationByName(self, name):
        """Return a named content item."""
        for item in self.operations:
            if item.name == name:
                return item
        raise KeyError("No operation named %s" % name)

    def get_schema(self, typename):
        """
        Get layer schema compatible with :class:`fiona` schema object
        """

        return get_schema(self.url, typename, self.version, auth=self.auth)


class ServiceIdentification(object):
    """ Implements IServiceIdentificationMetadata """

    def __init__(self, infoset, version):
        self._root = infoset
        self.type = testXMLValue(self._root.find(nspath("Name")))
        self.version = version
        self.title = testXMLValue(self._root.find(nspath("Title")))
        self.abstract = testXMLValue(self._root.find(nspath("Abstract")))
        self.keywords = [f.text for f in self._root.findall(nspath("Keywords"))]
        self.fees = testXMLValue(self._root.find(nspath("Fees")))
        self.accessconstraints = testXMLValue(
            self._root.find(nspath("AccessConstraints"))
        )


class ServiceProvider(object):
    """ Implements IServiceProviderMetatdata """

    def __init__(self, infoset):
        self._root = infoset
        self.name = testXMLValue(self._root.find(nspath("Name")))
        self.url = testXMLValue(self._root.find(nspath("OnlineResource")))
        self.keywords = extract_xml_list(self._root.find(nspath("Keywords")))


class ContentMetadata(AbstractContentMetadata):
    """Abstraction for WFS metadata.

    Implements IMetadata.
    """

    def __init__(
        self, elem, parent, parse_remote_metadata=False, timeout=30, auth=None
    ):
        """."""
        super(ContentMetadata, self).__init__(auth)
        self.id = testXMLValue(elem.find(nspath("Name")))
        self.title = testXMLValue(elem.find(nspath("Title")))
        self.abstract = testXMLValue(elem.find(nspath("Abstract")))
        self.keywords = [f.text for f in elem.findall(nspath("Keywords"))]

        # bboxes
        self.boundingBox = None
        b = elem.find(nspath("LatLongBoundingBox"))
        srs = elem.find(nspath("SRS"))

        if b is not None:
            self.boundingBox = (
                float(b.attrib["minx"]),
                float(b.attrib["miny"]),
                float(b.attrib["maxx"]),
                float(b.attrib["maxy"]),
                Crs(srs.text),
            )

        # transform wgs84 bbox from given default bboxt
        self.boundingBoxWGS84 = None

        if b is not None and srs is not None:
            wgs84 = pyproj.Proj("epsg:4326")
            try:
                src_srs = pyproj.Proj(srs.text)
                mincorner = pyproj.transform(
                    src_srs, wgs84, b.attrib["minx"], b.attrib["miny"],
                    always_xy=True,
                )
                maxcorner = pyproj.transform(
                    src_srs, wgs84, b.attrib["maxx"], b.attrib["maxy"],
                    always_xy=True,
                )

                self.boundingBoxWGS84 = (
                    mincorner[0],
                    mincorner[1],
                    maxcorner[0],
                    maxcorner[1],
                )
            except RuntimeError:
                pass

        # crs options
        self.crsOptions = [Crs(srs.text) for srs in elem.findall(nspath("SRS"))]

        # verbs
        self.verbOptions = [op.tag for op in parent.findall(nspath("Operations/*"))]
        self.verbOptions + [
            op.tag
            for op in elem.findall(nspath("Operations/*"))
            if op.tag not in self.verbOptions
        ]

        # others not used but needed for iContentMetadata harmonisation
        self.styles = None
        self.timepositions = None
        self.defaulttimeposition = None

        # MetadataURLs
        self.metadataUrls = []
        for m in elem.findall(nspath("MetadataURL")):
            metadataUrl = {
                "type": testXMLValue(m.attrib["type"], attrib=True),
                "format": testXMLValue(m.attrib["format"], attrib=True),
                "url": testXMLValue(m),
            }
            self.metadataUrls.append(metadataUrl)

    def parse_remote_metadata(self, timeout=30):
        """Parse remote metadata for MetadataURL of format 'XML' and add it as metadataUrl['metadata']"""
        for metadataUrl in self.metadataUrls:
            if (
                metadataUrl["url"] is not None and metadataUrl["format"].lower() == "xml"
            ):
                try:
                    content = openURL(
                        metadataUrl["url"], timeout=timeout, headers=self.headers, auth=self.auth
                    )
                    doc = etree.fromstring(content.read())
                    if metadataUrl["type"] == "FGDC":
                        mdelem = doc.find(".//metadata")
                        if mdelem is not None:
                            metadataUrl["metadata"] = Metadata(mdelem)
                        else:
                            metadataUrl["metadata"] = None
                    elif metadataUrl["type"] == "TC211":
                        mdelem = doc.find(
                            ".//" + util.nspath_eval("gmd:MD_Metadata", n.get_namespaces(["gmd"]))
                        ) or doc.find(
                            ".//" + util.nspath_eval("gmi:MI_Metadata", n.get_namespaces(["gmi"]))
                        )
                        if mdelem is not None:
                            metadataUrl["metadata"] = MD_Metadata(mdelem)
                        else:
                            metadataUrl["metadata"] = None
                except Exception:
                    metadataUrl["metadata"] = None


class OperationMetadata:
    """Abstraction for WFS metadata.

    Implements IMetadata.
    """

    def __init__(self, elem):
        """."""
        self.name = xmltag_split(elem.tag)
        # formatOptions
        self.formatOptions = [f.tag for f in elem.findall(nspath("ResultFormat/*"))]
        self.methods = []
        for verb in elem.findall(nspath("DCPType/HTTP/*")):
            url = verb.attrib["onlineResource"]
            self.methods.append({"type": xmltag_split(verb.tag), "url": url})



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2009 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

"""
API for OGC Filter Encoding (FE) constructs and metadata.

Filter Encoding: http://www.opengeospatial.org/standards/filter

Currently supports version 1.1.0 (04-095).
"""

from owslib.etree import etree
from owslib import util
from owslib.namespaces import Namespaces


# default variables
def get_namespaces():
    n = Namespaces()
    ns = n.get_namespaces(["dif", "fes", "gml", "ogc", "xs", "xsi"])
    ns[None] = n.get_namespace("ogc")
    return ns


namespaces = get_namespaces()
schema = 'http://schemas.opengis.net/filter/1.1.0/filter.xsd'
schema_location = '%s %s' % (namespaces['ogc'], schema)


class FilterRequest(object):
    """ filter class """
    def __init__(self, parent=None, version='1.1.0'):
        """

        filter Constructor

        Parameters
        ----------

        - parent: parent etree.Element object (default is None)
        - version: version (default is '1.1.0')

        """

        self.version = version
        self._root = etree.Element(util.nspath_eval('ogc:Filter', namespaces))
        if parent is not None:
            self._root.set(util.nspath_eval('xsi:schemaLocation', namespaces), schema_location)

    def set(self, parent=False, qtype=None, keywords=[], typenames='csw:Record', propertyname='csw:AnyText', bbox=None,
            identifier=None):
        """

        Construct and process a  GetRecords request

        Parameters
        ----------

        - parent: the parent Element object.  If this is not, then generate a standalone request
        - qtype: type of resource to query (i.e. service, dataset)
        - keywords: list of keywords
        - propertyname: the PropertyName to Filter against
        - bbox: the bounding box of the spatial query in the form [minx,miny,maxx,maxy]
        - identifier: the dc:identifier to query against with a PropertyIsEqualTo.  Ignores all other inputs.

        """

        # Set the identifier if passed.  Ignore other parameters
        dc_identifier_equals_filter = None
        if identifier is not None:
            dc_identifier_equals_filter = PropertyIsEqualTo('dc:identifier', identifier)
            self._root.append(dc_identifier_equals_filter.toXML())
            return self._root

        # Set the query type if passed
        dc_type_equals_filter = None
        if qtype is not None:
            dc_type_equals_filter = PropertyIsEqualTo('dc:type', qtype)

        # Set a bbox query if passed
        bbox_filter = None
        if bbox is not None:
            bbox_filter = BBox(bbox)

        # Set a keyword query if passed
        keyword_filter = None
        if len(keywords) > 0:
            if len(keywords) > 1:  # loop multiple keywords into an Or
                ks = []
                for i in keywords:
                    ks.append(PropertyIsLike(propertyname, "*%s*" % i, wildCard="*"))
                keyword_filter = Or(operations=ks)
            elif len(keywords) == 1:  # one keyword
                keyword_filter = PropertyIsLike(propertyname, "*%s*" % keywords[0], wildCard="*")

        # And together filters if more than one exists
        filters = [_f for _f in [keyword_filter, bbox_filter, dc_type_equals_filter] if _f]
        if len(filters) == 1:
            self._root.append(filters[0].toXML())
        elif len(filters) > 1:
            self._root.append(And(operations=filters).toXML())

        return self._root

    def setConstraint(self, constraint, tostring=False):
        """
        Construct and process a  GetRecords request

        Parameters
        ----------

        - constraint: An OgcExpression object
        - tostring (optional): return as string

        """
        self._root.append(constraint.toXML())

        if tostring:
            return util.element_to_string(self._root, xml_declaration=False)

        return self._root

    def setConstraintList(self, constraints, tostring=False):
        """
        Construct and process a  GetRecords request

        Parameters
        ----------

        - constraints: A list of OgcExpression objects
                       The list is interpretted like so:

                       [a,b,c]
                       a || b || c

                       [[a,b,c]]
                       a && b && c

                       [[a,b],[c],[d],[e]] or [[a,b],c,d,e]
                       (a && b) || c || d || e
        - tostring (optional): return as string

        """
        ors = []
        if len(constraints) == 1:
            if isinstance(constraints[0], OgcExpression):
                flt = self.setConstraint(constraints[0])
            else:
                self._root.append(And(operations=constraints[0]).toXML())
                flt = self._root
            if tostring:
                return util.element_to_string(flt, xml_declaration=False)
            else:
                return flt

        for c in constraints:
            if isinstance(c, OgcExpression):
                ors.append(c)
            elif isinstance(c, list) or isinstance(c, tuple):
                if len(c) == 1:
                    ors.append(c[0])
                elif len(c) >= 2:
                    ands = []
                    for sub in c:
                        if isinstance(sub, OgcExpression):
                            ands.append(sub)
                    ors.append(And(operations=ands))

        self._root.append(Or(operations=ors).toXML())

        if tostring:
            return util.element_to_string(self._root, xml_declaration=False)

        return self._root


class FilterCapabilities(object):
    """ Abstraction for Filter_Capabilities """
    def __init__(self, elem):
        # Spatial_Capabilities
        self.spatial_operands = [f.text for f in elem.findall(util.nspath_eval(
            'ogc:Spatial_Capabilities/ogc:GeometryOperands/ogc:GeometryOperand', namespaces))]
        self.spatial_operators = []
        for f in elem.findall(util.nspath_eval(
                'ogc:Spatial_Capabilities/ogc:SpatialOperators/ogc:SpatialOperator', namespaces)):
            self.spatial_operators.append(f.attrib['name'])

        # Temporal_Capabilities
        self.temporal_operands = [f.text for f in elem.findall(util.nspath_eval(
            'ogc:Temporal_Capabilities/ogc:TemporalOperands/ogc:TemporalOperand', namespaces))]
        self.temporal_operators = []
        for f in elem.findall(util.nspath_eval(
                'ogc:Temporal_Capabilities/ogc:TemporalOperators/ogc:TemporalOperator', namespaces)):
            self.temporal_operators.append(f.attrib['name'])

        # Scalar_Capabilities
        self.scalar_comparison_operators = [f.text for f in elem.findall(util.nspath_eval(
            'ogc:Scalar_Capabilities/ogc:ComparisonOperators/ogc:ComparisonOperator', namespaces))]


class FilterCapabilities200(object):
    """Abstraction for Filter_Capabilities 2.0"""
    def __init__(self, elem):
        # Spatial_Capabilities
        self.spatial_operands = [f.attrib.get('name') for f in elem.findall(util.nspath_eval(
            'fes:Spatial_Capabilities/fes:GeometryOperands/fes:GeometryOperand', namespaces))]
        self.spatial_operators = []
        for f in elem.findall(util.nspath_eval(
                'fes:Spatial_Capabilities/fes:SpatialOperators/fes:SpatialOperator', namespaces)):
            self.spatial_operators.append(f.attrib['name'])

        # Temporal_Capabilities
        self.temporal_operands = [f.attrib.get('name') for f in elem.findall(util.nspath_eval(
            'fes:Temporal_Capabilities/fes:TemporalOperands/fes:TemporalOperand', namespaces))]
        self.temporal_operators = []
        for f in elem.findall(util.nspath_eval(
                'fes:Temporal_Capabilities/fes:TemporalOperators/fes:TemporalOperator', namespaces)):
            self.temporal_operators.append(f.attrib['name'])

        # Scalar_Capabilities
        self.scalar_comparison_operators = [f.text for f in elem.findall(util.nspath_eval(
            'fes:Scalar_Capabilities/fes:ComparisonOperators/fes:ComparisonOperator', namespaces))]

        # Conformance
        self.conformance = []
        for f in elem.findall(util.nspath_eval('fes:Conformance/fes:Constraint', namespaces)):
            self.conformance[f.attrib.get('name')] = f.find(util.nspath_eval('fes:DefaultValue', namespaces)).text


def setsortby(parent, propertyname, order='ASC'):
    """

    constructs a SortBy element

    Parameters
    ----------

    - parent: parent etree.Element object
    - propertyname: the PropertyName
    - order: the SortOrder (default is 'ASC')

    """

    tmp = etree.SubElement(parent, util.nspath_eval('ogc:SortBy', namespaces))
    tmp2 = etree.SubElement(tmp, util.nspath_eval('ogc:SortProperty', namespaces))
    etree.SubElement(tmp2, util.nspath_eval('ogc:PropertyName', namespaces)).text = propertyname
    etree.SubElement(tmp2, util.nspath_eval('ogc:SortOrder', namespaces)).text = order


class SortProperty(object):
    def __init__(self, propertyname, order='ASC'):
        self.propertyname = propertyname
        self.order = order.upper()
        if self.order not in ['DESC', 'ASC']:
            raise ValueError("SortOrder can only be 'ASC' or 'DESC'")

    def toXML(self):
        node0 = etree.Element(util.nspath_eval("ogc:SortProperty", namespaces))
        etree.SubElement(node0, util.nspath_eval('ogc:PropertyName', namespaces)).text = self.propertyname
        etree.SubElement(node0, util.nspath_eval('ogc:SortOrder', namespaces)).text = self.order
        return node0


class SortBy(object):
    def __init__(self, properties):
        self.properties = properties

    def toXML(self):
        node0 = etree.Element(util.nspath_eval("ogc:SortBy", namespaces))
        for prop in self.properties:
            node0.append(prop.toXML())
        return node0


class OgcExpression(object):
    def __init__(self):
        pass


class BinaryComparisonOpType(OgcExpression):
    """ Super class of all the property operation classes"""
    def __init__(self, propertyoperator, propertyname, literal, matchcase=True):
        self.propertyoperator = propertyoperator
        self.propertyname = propertyname
        self.literal = literal
        self.matchcase = matchcase

    def toXML(self):
        node0 = etree.Element(util.nspath_eval(self.propertyoperator, namespaces))
        if not self.matchcase:
            node0.set('matchCase', 'false')
        etree.SubElement(node0, util.nspath_eval('ogc:PropertyName', namespaces)).text = self.propertyname
        etree.SubElement(node0, util.nspath_eval('ogc:Literal', namespaces)).text = self.literal
        return node0


class PropertyIsEqualTo(BinaryComparisonOpType):
    """ PropertyIsEqualTo class"""
    def __init__(self, propertyname, literal, matchcase=True):
        BinaryComparisonOpType.__init__(self, 'ogc:PropertyIsEqualTo', propertyname, literal, matchcase)


class PropertyIsNotEqualTo(BinaryComparisonOpType):
    """ PropertyIsNotEqualTo class """
    def __init__(self, propertyname, literal, matchcase=True):
        BinaryComparisonOpType.__init__(self, 'ogc:PropertyIsNotEqualTo', propertyname, literal, matchcase)


class PropertyIsLessThan(BinaryComparisonOpType):
    """PropertyIsLessThan class"""
    def __init__(self, propertyname, literal, matchcase=True):
        BinaryComparisonOpType.__init__(self, 'ogc:PropertyIsLessThan', propertyname, literal, matchcase)


class PropertyIsGreaterThan(BinaryComparisonOpType):
    """PropertyIsGreaterThan class"""
    def __init__(self, propertyname, literal, matchcase=True):
        BinaryComparisonOpType.__init__(self, 'ogc:PropertyIsGreaterThan', propertyname, literal, matchcase)


class PropertyIsLessThanOrEqualTo(BinaryComparisonOpType):
    """PropertyIsLessThanOrEqualTo class"""
    def __init__(self, propertyname, literal, matchcase=True):
        BinaryComparisonOpType.__init__(self, 'ogc:PropertyIsLessThanOrEqualTo', propertyname, literal, matchcase)


class PropertyIsGreaterThanOrEqualTo(BinaryComparisonOpType):
    """PropertyIsGreaterThanOrEqualTo class"""
    def __init__(self, propertyname, literal, matchcase=True):
        BinaryComparisonOpType.__init__(self, 'ogc:PropertyIsGreaterThanOrEqualTo', propertyname, literal, matchcase)


class PropertyIsLike(OgcExpression):
    """PropertyIsLike class"""
    def __init__(self, propertyname, literal, escapeChar='\\', singleChar='_', wildCard='%', matchCase=True):
        self.propertyname = propertyname
        self.literal = literal
        self.escapeChar = escapeChar
        self.singleChar = singleChar
        self.wildCard = wildCard
        self.matchCase = matchCase

    def toXML(self):
        node0 = etree.Element(util.nspath_eval('ogc:PropertyIsLike', namespaces))
        node0.set('wildCard', self.wildCard)
        node0.set('singleChar', self.singleChar)
        node0.set('escapeChar', self.escapeChar)
        if not self.matchCase:
            node0.set('matchCase', 'false')
        etree.SubElement(node0, util.nspath_eval('ogc:PropertyName', namespaces)).text = self.propertyname
        etree.SubElement(node0, util.nspath_eval('ogc:Literal', namespaces)).text = self.literal
        return node0


class PropertyIsNull(OgcExpression):
    """PropertyIsNull class"""
    def __init__(self, propertyname):
        self.propertyname = propertyname

    def toXML(self):
        node0 = etree.Element(util.nspath_eval('ogc:PropertyIsNull', namespaces))
        etree.SubElement(node0, util.nspath_eval('ogc:PropertyName', namespaces)).text = self.propertyname
        return node0


class PropertyIsBetween(OgcExpression):
    """PropertyIsBetween class"""
    def __init__(self, propertyname, lower, upper):
        self.propertyname = propertyname
        self.lower = lower
        self.upper = upper

    def toXML(self):
        node0 = etree.Element(util.nspath_eval('ogc:PropertyIsBetween', namespaces))
        etree.SubElement(node0, util.nspath_eval('ogc:PropertyName', namespaces)).text = self.propertyname
        node1 = etree.SubElement(node0, util.nspath_eval('ogc:LowerBoundary', namespaces))
        etree.SubElement(node1, util.nspath_eval('ogc:Literal', namespaces)).text = '%s' % self.lower
        node2 = etree.SubElement(node0, util.nspath_eval('ogc:UpperBoundary', namespaces))
        etree.SubElement(node2, util.nspath_eval('ogc:Literal', namespaces)).text = '%s' % self.upper
        return node0


class BBox(OgcExpression):
    """Construct a BBox, two pairs of coordinates (west-south and east-north)"""
    def __init__(self, bbox, crs=None):
        self.bbox = bbox
        self.crs = crs

    def toXML(self):
        tmp = etree.Element(util.nspath_eval('ogc:BBOX', namespaces))
        etree.SubElement(tmp, util.nspath_eval('ogc:PropertyName', namespaces)).text = 'ows:BoundingBox'
        tmp2 = etree.SubElement(tmp, util.nspath_eval('gml:Envelope', namespaces))
        if self.crs is not None:
            tmp2.set('srsName', self.crs)
        etree.SubElement(tmp2, util.nspath_eval('gml:lowerCorner', namespaces)).text = '{} {}'.format(
            self.bbox[0], self.bbox[1])
        etree.SubElement(tmp2, util.nspath_eval('gml:upperCorner', namespaces)).text = '{} {}'.format(
            self.bbox[2], self.bbox[3])
        return tmp


# BINARY
class BinaryLogicOpType(OgcExpression):
    """ Binary Operators: And / Or """
    def __init__(self, binary_operator, operations):
        self.binary_operator = binary_operator
        try:
            assert len(operations) >= 2
            self.operations = operations
        except Exception:
            raise ValueError("Binary operations (And / Or) require a minimum of two operations to operate against")

    def toXML(self):
        node0 = etree.Element(util.nspath_eval(self.binary_operator, namespaces))
        for op in self.operations:
            node0.append(op.toXML())
        return node0


class And(BinaryLogicOpType):
    def __init__(self, operations):
        super(And, self).__init__('ogc:And', operations)


class Or(BinaryLogicOpType):
    def __init__(self, operations):
        super(Or, self).__init__('ogc:Or', operations)


# UNARY
class UnaryLogicOpType(OgcExpression):
    """ Unary Operator: Not """
    def __init__(self, urary_operator, operations):
        self.urary_operator = urary_operator
        self.operations = operations

    def toXML(self):
        node0 = etree.Element(util.nspath_eval(self.urary_operator, namespaces))
        for op in self.operations:
            node0.append(op.toXML())
        return node0


class Not(UnaryLogicOpType):
    def __init__(self, operations):
        super(Not, self).__init__('ogc:Not', operations)



# -*- coding: utf-8 -*-
# =============================================================================
# Authors : Alexander Kmoch <allixender@gmail.com>
#
# =============================================================================

"""
API for OGC Web Services Context Document (OWS Context) format.

Conceptual model base classes: http://www.opengeospatial.org/standards/owc

OGC OWS Context Conceptual Model 1.0 (12-080r2)
"""

from owslib.owscontext.atom import decode_atomxml, encode_atomxml
from owslib.owscontext.common import GENERIC_OWCSPEC_URL
# from owslib.util import log
# TODO make dates from (currently) string to datetime instances
from owslib.owscontext.common import TimeIntervalFormat
from owslib.owscontext.common import try_float, try_int, \
    extract_p, build_from_xp
from owslib.owscontext.geojson import encode_json, decode_json


class OwcContext(object):
    """
    * + specReference :URI
    * + language :CharacterString
    * + id :CharacterString
    * + title :CharacterString
    * + abstract :CharacterString [0..1]
    * + updateDate :CharacterString [0..1]
    * + author :CharacterString [0..*]
    * + publisher :CharacterString [0..1]
    * + creator :Creator [0..1]
    * +----+ creatorApplication :CreatorApplication [0..1]
    * +----+ creatorDisplay :CreatorDisplay [0..1]
    * + rights :CharacterString [0..1]
    * + areaOfInterest :GM_Envelope [0..1]
    * + timeIntervalOfInterest :TM_GeometricPrimitive [0..1]
    * + keyword :CharacterString [0..*]
    * + extension :Any [0..*]
    """

    def __init__(self,
                 id,
                 update_date,
                 title,
                 language="en",
                 spec_reference=[],
                 area_of_interest=[],
                 context_metadata=[],
                 subtitle=None,
                 authors=[],
                 publisher=None,
                 creator_application=None,
                 creator_display=None,
                 rights=None,
                 time_interval_of_interest=None,
                 keywords=[],
                 resources=[]
                 ):
        """
        constructor:

        :param id: URL
        :param update_date: datetime
        :param title:   String,
        :param language: String = "en"
        :param spec_reference: OwcLink[]  links.profiles[] and rel=profile
        :param area_of_interest: Option[Rectangle] = None, aka Double[]
        :param context_metadata: OwcLink[] links.via[] and rel=via
        :param subtitle: Option[String] = None,
        :param authors: List[OwcAuthor]
        :param publisher: Option[String] = None,
        :param creator_application: Option[OwcCreatorApplication] = None,
        :param creator_display: Option[OwcCreatorDisplay] = None,
        :param rights: Option[String] = None,
        :param time_interval_of_interest: TimeIntervalFormat(start,end)
        :param keywords: OwcCategory[]
        :param resources: OwcResource[]
        """
        self.id = id
        self.spec_reference = spec_reference
        self.area_of_interest = area_of_interest
        self.context_metadata = context_metadata
        self.language = language
        self.title = title
        self.subtitle = subtitle
        self.update_date = update_date
        self.authors = authors
        self.publisher = publisher
        self.creator_application = creator_application
        self.creator_display = creator_display
        self.rights = rights
        self.time_interval_of_interest = time_interval_of_interest
        self.keywords = keywords
        self.resources = resources

        # TODO spec reference, check or provide?
        if len(self.spec_reference) <= 0:
            self.spec_reference.append(
                OwcLink(href=GENERIC_OWCSPEC_URL, rel='profile'))

        # TODO check and validate? how much?

    def to_dict(self):
        """

        :return: dict of obj
        """
        return {
            "type": "FeatureCollection",
            "id": self.id,
            "bbox": self.area_of_interest,
            "properties": {
                "lang": self.language,
                "title": self.title,
                "abstract": self.subtitle,
                "updated": self.update_date,
                "date":
                    None if self.time_interval_of_interest is None else
                    self.time_interval_of_interest.__str__(),
                "rights": self.rights,
                "authors":
                    [] if len(self.authors) <= 0 else
                    [obj.to_dict() for obj in self.authors],
                "publisher": self.publisher,
                "display":
                    None if self.creator_display is None else
                    self.creator_display.to_dict(),
                "generator":
                    None if self.creator_application is None else
                    self.creator_application.to_dict(),
                "categories":
                    [] if len(self.keywords) <= 0 else
                    [obj.to_dict() for obj in self.keywords],
                "links": {
                    "profiles":
                        [] if len(self.spec_reference) <= 0 else
                        [obj.to_dict() for obj in self.spec_reference],
                    "via":
                        [] if len(self.context_metadata) <= 0 else
                        [obj.to_dict() for obj in self.context_metadata]},
            },
            "features":
                [] if len(self.resources) <= 0 else
                [obj.to_dict() for obj in self.resources]
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        # TODO add spec url conversion from generic to geojson
        return encode_json(self.to_dict())

    def to_atomxml(self):
        # TODO add spec url conversion from generic to atom
        return encode_atomxml(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        # TODO parse bbox??
        return OwcContext(
            id=d['id'],
            spec_reference=[OwcLink.from_dict(do) for do in
                            extract_p('properties.links.profiles', d, [])],
            area_of_interest=extract_p('bbox', d, None),
            context_metadata=[OwcLink.from_dict(do) for do in
                              extract_p('properties.links.via', d, [])],
            language=extract_p('properties.lang', d, None),
            title=extract_p('properties.title', d, None),
            subtitle=extract_p('properties.abstract', d, None),
            update_date=extract_p('properties.updated', d, None),
            authors=[OwcAuthor.from_dict(do) for do in
                     extract_p('properties.authors', d, [])],
            publisher=extract_p('properties.publisher', d, None),
            creator_application=build_from_xp(
                'properties.generator', d, OwcCreatorApplication, None),
            creator_display=build_from_xp(
                'properties.display', d, OwcCreatorDisplay, None),
            rights=extract_p('properties.rights', d, None),
            time_interval_of_interest=TimeIntervalFormat.from_string(
                extract_p('properties.date', d, None)),
            keywords=[OwcCategory.from_dict(do) for do in
                      extract_p('properties.categories', d, [])],
            resources=[OwcResource.from_dict(do) for do in
                       extract_p('features', d, [])]
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        # TODO add spec url conversion from geojson to generic
        # TODO should validate if geojson Type == FeatureCollection?
        return cls.from_dict(d)

    @classmethod
    def from_atomxml(cls, xml_bytes):
        """
        lets see if we can reuse the dict structure builder from geojson
        :param xmlstring:
        :return: OwcContext
        """
        # TODO add spec url conversion from atom to generic
        # TODO should validate anything?
        # TODO ValueError: Unicode strings with encoding declaration
        # are not supported. Please use bytes input or XML fragments
        # without declaration -> Decide, currently bytes
        d = decode_atomxml(xml_bytes)
        return cls.from_dict(d)


class OwcResource(object):
    """
    * + id :CharacterString -
        Unambiguous reference to the identification of the resource,
        SHALL contain an (IRI) URI value
    * + title :CharacterString
    * + abstract :CharacterString [0..1]
    * + updateDate :TM_Date [0..1]
    * + author :CharacterString [0..*]
    * + publisher :CharacterString [0..1]
    * + rights :CharacterString [0..1]
    * + geospatialExtent :GM_Envelope [0..1]
    * + temporalExtent :TM_GeometricPrimitive [0..1]
    * + contentDescription :Any [0..1] // aka alternates
    * + preview :URI [0..*]
    * + contentByRef :URI [0..*]
    * + offering :Offering [0..*]
    * + active :Boolean [0..1]
    * + keyword :CharacterString [0..*]
    * + maxScaleDenominator :double [0..1]
    * + minScaleDenominator :double [0..1]
    * + folder :CharacterString [0..1]
    * + extension :Any [0..*]
    """

    def __init__(self,
                 id,
                 title,
                 update_date,
                 subtitle=None,
                 authors=[],
                 publisher=None,
                 rights=None,
                 geospatial_extent=None,
                 temporal_extent=None,
                 content_description=[],
                 preview=[],
                 content_by_ref=[],
                 offerings=[],
                 active=False,
                 resource_metadata=[],
                 keywords=[],
                 min_scale_denominator=None,
                 max_scale_denominator=None,
                 folder=None
                 ):
        """
        constructor:

        :param id: URL
        :param title: String
        :param update_date: datetime
        :param subtitle: String
        :param authors: List[OwcAuthor]
        :param publisher: String
        :param rights: String
        :param geospatial_extent: currently GeoJSON Polygon String
        :param temporal_extent: str
        :param content_description: OwcLink[] links.alternates, rel=alternate
        :param preview: OwcLink[] aka links.previews[] and rel=icon (atom)
        :param content_by_ref: OwcLink[], links.data, rel=enclosure (atom)
        :param offerings: OwcOffering[]
        :param active: Boolean
        :param resource_metadata: OwcLink[] aka links.via[] & rel=via
        :param keywords: OwcCategory[]
        :param min_scale_denominator: Double
        :param max_scale_denominator: Double
        :param folder: String
        """
        # # TimeIntervalFormat(start,end)
        self.id = id
        self.title = title
        self.subtitle = subtitle
        self.update_date = update_date
        self.authors = authors
        self.publisher = publisher
        self.rights = rights
        self.geospatial_extent = geospatial_extent
        self.temporal_extent = temporal_extent
        self.content_description = content_description
        self.preview = preview
        self.content_by_ref = content_by_ref
        self.offerings = offerings
        self.active = active
        self.resource_metadata = resource_metadata
        self.keywords = keywords
        self.min_scale_denominator = min_scale_denominator
        self.max_scale_denominator = max_scale_denominator
        self.folder = folder

    def to_dict(self):
        # TODO parse geometry??
        return {
            "type": "Feature",
            "id": self.id,
            "geometry": self.geospatial_extent,
            "properties": {
                "title": self.title,
                "abstract": self.subtitle,
                "updated": self.update_date,
                "date":
                    None if self.temporal_extent is None else
                    self.temporal_extent.__str__(),
                "authors":
                    [] if len(self.authors) <= 0 else
                    [obj.to_dict() for obj in self.authors],
                "publisher": self.publisher,
                "rights": self.rights,
                "categories":
                    [] if len(self.keywords) <= 0 else
                    [obj.to_dict() for obj in self.keywords],
                "links": {
                    "alternates":
                        [] if len(self.content_description) <= 0 else
                        [obj.to_dict() for obj in
                         self.content_description],
                    "previews":
                        [] if len(self.preview) <= 0 else
                        [obj.to_dict() for obj in self.preview],
                    "data":
                        [] if len(self.content_by_ref) <= 0 else
                        [obj.to_dict() for obj in
                         self.content_by_ref],
                    "via":
                        [] if len(self.resource_metadata) <= 0 else
                        [obj.to_dict() for obj in
                         self.resource_metadata]},
                "offerings":
                    [] if len(self.offerings) <= 0 else
                    [obj.to_dict() for obj in self.offerings],
                "active": self.active,
                "minscaledenominator": self.min_scale_denominator,
                "maxscaledenominator": self.max_scale_denominator,
                "folder": self.folder
            }
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcResource(
            id=d['id'],
            geospatial_extent=extract_p('geometry', d, None),
            title=d['properties']['title'],
            subtitle=extract_p('properties.abstract', d, None),
            update_date=extract_p('properties.updated', d, None),
            authors=[OwcAuthor.from_dict(do) for do in
                     extract_p('properties.authors', d, [])],
            publisher=extract_p('properties.publisher', d, None),
            rights=extract_p('properties.rights', d, None),
            temporal_extent=TimeIntervalFormat.from_string(
                extract_p('properties.date', d, None)),
            keywords=[OwcCategory.from_dict(do) for do in
                      extract_p('properties.categories', d, [])],
            resource_metadata=[OwcLink.from_dict(do) for do in
                               extract_p('properties.links.via', d, [])],
            content_description=[OwcLink.from_dict(do)
                                 for do in extract_p('properties.links.alternates', d, [])],
            preview=[OwcLink.from_dict(do) for do in
                     extract_p('properties.links.previews', d, [])],
            content_by_ref=[OwcLink.from_dict(do) for do in
                            extract_p('properties.links.data', d, [])],
            offerings=[OwcOffering.from_dict(do) for do in
                       extract_p('properties.offerings', d, [])],
            active=extract_p('properties.active', d, None),
            min_scale_denominator=try_float(extract_p(
                'properties.minscaledenominator', d, None)),
            max_scale_denominator=try_float(extract_p(
                'properties.maxscaledenominator', d, None)),
            folder=extract_p('properties.folder', d, None),
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        # TODO should validate if geojson Type == Feature?
        return cls.from_dict(d)


class OwcCreator(object):
    """
    * + creatorApplication :CreatorApplication [0..1]
    * + creatorDisplay :CreatorDisplay [0..1]
    * + extension :Any [0..*]
    *
    * OWCCreator base class is never realized by itself,
        neither in Atom nor in GeoJSON Encoding;
    * Creator/Display and Creator/Application are instantiated as completely
        separate entities in both Atom and GeoJSON
    """

    def __init__(self,
                 creator_application=None,
                 creator_display=None
                 ):
        """
        constructor:

        :param creator_application: OwcCreatorApplication
        :param creator_display: OwcCreatorDisplay
        """
        self.creator_application = creator_application
        self.creator_display = creator_display


class OwcCreatorApplication(object):
    """
    CreatorApplication
    + title :CharacterString [0..1]
    + uri :URI [0..1]
    + version :Version [0..1]
    + the single only class that doesn't have explicit extension?
        (but would be inherited from OWC:Creator?)
    """

    def __init__(self,
                 title,
                 uri=None,
                 version=None
                 ):
        """
        constructor:

        :param title: String
        :param uri: URL
        :param version: String
        """
        self.title = title
        self.uri = uri
        self.version = version

    def to_dict(self):
        return {
            "title": self.title,
            "uri": self.uri,
            "version": self.version
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcCreatorApplication(
            title=extract_p('title', d, None),
            uri=extract_p('uri', d, None),
            version=extract_p('version', d, None)
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcCreatorDisplay(object):
    """
    CreatorDisplay
    + pixelWidth :int [0..1]
    + pixelHeight :int [0..1]
    + mmPerPixel :double [0..1]
    + extension :Any [0..*]
    """

    def __init__(self,
                 pixel_width,
                 pixel_height=None,
                 mm_per_pixel=None,
                 ):
        """
        constructor:

        :param pixel_width: Double
        :param pixel_height: Double
        :param mm_per_pixel: Double
        """
        self.pixel_width = pixel_width
        self.pixel_height = pixel_height
        self.mm_per_pixel = mm_per_pixel

    def to_dict(self):
        return {
            "pixelWidth": self.pixel_width,
            "pixelHeight": self.pixel_height,
            "mmPerPixel": self.mm_per_pixel
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcCreatorDisplay(
            pixel_width=try_int(extract_p('pixelWidth', d, None)),
            pixel_height=try_int(extract_p('pixelHeight', d, None)),
            mm_per_pixel=try_float(extract_p('mmPerPixel', d, None))
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcLink(object):
    """
    * OwcLink in most cases will have an array of links at the path of the rel
    *
    * OWC:Context specReference attribute:
        atom rel="profile" - geojson links.profiles[] array
    * OWC:Context contextMetadata attribute:
        atom rel="via" - geojson links.via[] array
    *
    * OWC:Resource contentDescription attribute:
        atom rel="alternate" - geojson links.alternates[] array
    * OWC:Resource preview attribute:
        atom rel="icon" - geojson links.previews[] array
    * OWC:Resource contentByRef attribute:
        atom rel="enclosure" - geojson links.data[] array
    * OWC:Resource resourceMetadata attribute:
        atom rel="via" - geojson links.via[] array
    *
    * links for data and previews
        (aka rels enclosure and icon should have length attributes set)
    *
    * + href :URI
    * + type :String [0..1]
    * + lang :String [0..1]
    * + title :String [0..1]
    * + length :Integer [0..1]
    * + extension :Any [0..*]
    """

    def __init__(self,
                 href,
                 rel,
                 mimetype=None,
                 lang=None,
                 title=None,
                 length=None,
                 ):
        """
        constructor:

        :param href: URL
        :param rel: String
        :param mimetype: String
        :param lang: String
        :param title: String
        :param length: Int
        """
        self.href = href
        self.rel = rel
        self.mimetype = mimetype
        self.lang = lang
        self.title = title
        self.length = length

    def to_dict(self):
        return {
            "href": self.href,
            "type": self.mimetype,
            "length": self.length,
            "lang": self.lang,
            "title": self.title,
            "rel": self.rel
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcLink(
            href=extract_p('href', d, None),
            rel=extract_p('rel', d, None),
            mimetype=extract_p('type', d, None),
            lang=extract_p('lang', d, None),
            title=extract_p('title', d, None),
            length=try_int(extract_p('length', d, None))
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcCategory(object):
    """
    keywords class
    """

    def __init__(self,
                 term,
                 scheme=None,
                 label=None,
                 ):
        """
        constructor:

        :param term: String
        :param scheme: String (can point to a controlled list, too, I guess)
        :param label: String
        """
        self.term = term
        self.scheme = scheme
        self.label = label

    def to_dict(self):
        return {
            "scheme": self.scheme,
            "term": self.term,
            "label": self.label
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcCategory(
            term=extract_p('term', d, None),
            scheme=extract_p('scheme', d, None),
            label=extract_p('label', d, None)
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcAuthor(object):
    """
    apparently handled differently in Atom and GeoJSON encodings?
    """

    def __init__(self,
                 name=None,
                 email=None,
                 uri=None
                 ):
        """
        constructor:

        :param name: String
        :param email: String (EmailAddress)
        :param uri: URL
        """
        self.name = name
        self.email = email
        self.uri = uri

    def to_dict(self):
        return {
            "name": self.name,
            "email": self.email,
            "uri": self.uri
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcAuthor(
            name=extract_p('name', d, None),
            email=extract_p('email', d, None),
            uri=extract_p('uri', d, None)
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcOffering(object):
    """
    * + code :URI
    * + operations :Offering [0..*]
    * + contents :Content [0..*]
    * + styles :StyleSet [0..*]
    * + extension :Any [0..*]
    """

    def __init__(self,
                 offering_code,
                 operations=[],
                 contents=[],
                 styles=[]
                 ):
        """
        constructor:

        :param offering_code: URL
        :param operations: List[OwcOperation]
        :param contents: List[OwcContent]
        :param styles: List[OwcStyleSet]
        """
        self.offering_code = offering_code
        self.operations = operations
        self.contents = contents
        self.styles = styles

    def to_dict(self):
        return {
            "code": self.offering_code,
            "operations": [] if len(self.operations) <= 0 else
            [obj.to_dict() for obj in self.operations],
            "contents": [] if len(self.contents) <= 0 else
            [obj.to_dict() for obj in self.contents],
            "styles": [] if len(self.styles) <= 0 else
            [obj.to_dict() for obj in self.styles]
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcOffering(
            offering_code=extract_p('code', d, None),
            operations=[OwcOperation.from_dict(do) for do in
                        extract_p('operations', d, [])],
            contents=[OwcContent.from_dict(do) for do in
                      extract_p('contents', d, [])],
            styles=[OwcStyleSet.from_dict(do) for do in
                    extract_p('styles', d, [])]
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcOperation(object):
    """
    * + code :CharacterString
    * + method :CharacterString
    * + type :CharacterString
    * + requestURL :URI
    * + request :Content [0..1]
    * + result :Any [0..1]
    * + extension :Any [0..*]
    """

    def __init__(self,
                 operations_code,
                 http_method,
                 request_url,
                 mimetype=None,
                 request=None,
                 result=None
                 ):
        """
        constructor:

        :param operations_code: String (e.g. GetCapabilities)
        :param http_method: String, HTTP verb (GET, POST ..)
        :param request_url: URL
        :param mimetype: String, MIME media type of the EXPECTED results
        :param request: Option[OwcContent] = None
        :param result:Option[OwcContent] = None
        """
        self.operations_code = operations_code
        self.http_method = http_method
        self.request_url = request_url
        self.mimetype = mimetype
        self.request = request
        self.result = result

    def to_dict(self):
        return {
            "code": self.operations_code,
            "method": self.http_method,
            "type": self.mimetype,
            "href": self.request_url,
            "request":
                None if self.request is None else self.request.to_dict(),
            "result":
                None if self.result is None else self.result.to_dict()
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcOperation(
            operations_code=extract_p('code', d, None),
            http_method=extract_p('method', d, None),
            mimetype=extract_p('type', d, None),
            request_url=extract_p('href', d, None),
            request=build_from_xp('request', d, OwcContent, None),
            result=build_from_xp('result', d, OwcContent, None)
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcStyleSet(object):
    """
    * + name :CharacterString
    * + title :CharacterString
    * + abstract :CharacterString [0..1]
    * + default :Boolean [0..1]
    * + legendURL :URI [0..*]
    * + content :Content [0..1]
    * + extension :Any [0..*]
    """

    def __init__(self,
                 name,
                 title,
                 subtitle=None,
                 is_default=None,
                 legend_url=None,
                 content=None
                 ):
        """
        constructor:

        :param name: String
        :param title: String
        :param subtitle: String
        :param is_default: Boolean
        :param legend_url: URL
        :param content: Option[OwcContent] = None
        """
        self.name = name
        self.title = title
        self.subtitle = subtitle
        self.is_default = is_default
        self.legend_url = legend_url
        self.content = content

    def to_dict(self):
        return {
            "name": self.name,
            "title": self.title,
            "abstract": self.subtitle,
            "default": self.is_default,
            "legendURL": self.legend_url,
            "content":
                None if self.content is None else self.content.to_dict()
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcStyleSet(
            name=extract_p('name', d, None),
            title=extract_p('title', d, None),
            subtitle=extract_p('abstract', d, None),
            is_default=extract_p('default', d, None),
            legend_url=extract_p('legendURL', d, None),
            content=build_from_xp('content', d, OwcContent, None)
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)


class OwcContent(object):
    """
    * + type :CharacterString
    * + URL :URI [0..1]
    * + content :Any [0..1]
    * + extension :Any [0..*]
    """

    def __init__(self,
                 mimetype,
                 content,
                 url=None,
                 title=None
                 ):
        """
        constructor:

        :param mimetype: String
        :param content: String encoded, actual content/data
        :param url: URL
        :param title: String
        """
        self.mimetype = mimetype
        self.content = content
        self.url = url
        self.title = title

    def to_dict(self):
        return {
            "type": self.mimetype,
            "url": self.url,
            "content": self.content,
            "title": self.title
        }

    def __str__(self):
        return str(self.to_dict())

    def to_json(self):
        return encode_json(self.to_dict())

    @classmethod
    def from_dict(cls, d):
        return OwcContent(
            mimetype=extract_p('type', d, None),
            content=extract_p('content', d, None),
            url=extract_p('url', d, None),
            title=extract_p('title', d, None)
        )

    @classmethod
    def from_json(cls, jsondata):
        d = decode_json(jsondata)
        return cls.from_dict(d)



from owslib.etree import etree
from owslib.util import nspath, testXMLValue, openURL
from owslib.util import xml_to_dict as _xml_to_dict
from datetime import datetime
from dateutil import parser

namespaces = {
    'wml1.1': '{http://www.cuahsi.org/waterML/1.1/}',
    'wml1.0': '{http://www.cuahsi.org/waterML/1.0/}',
    'xsi': '{http://www.w3.org/2001/XMLSchema-instance',
    'xsd': '{http://www.w3.org/2001/XMLSchema'
}


def ns(namespace):
    return namespaces.get(namespace)


class XMLParser(object):
    """
        Convienence class; provides some useful shortcut methods to make retrieving xml elements from etree
        a little easier.
    """
    def __init__(self, xml_root, namespace):
        try:
            self._root = etree.parse(xml_root)
        except Exception:
            self._root = xml_root

        if namespace not in namespaces:
            raise ValueError('Unsupported namespace passed in to parser!')

        self._ns = namespace

    def _find(self, tofind):
        try:
            return self._root.find(namespaces.get(self._ns) + tofind)
        except Exception:
            return None

    def _findall(self, tofind):
        try:
            return self._root.findall(namespaces.get(self._ns) + tofind)
        except Exception:
            return None


class SitesResponse(XMLParser):
    """
        Parses the response from a 'GetSites' request

        Parameters
        ===========
        :xmlio - A file-like object that holds the xml response from the request.

        Return
        =======
        An object constructed from a dictionary parse of the response. The object has get access and can iterate
        over the sites returned.
    """
    def __init__(self, xml, version='wml1.1'):
        super(SitesResponse, self).__init__(xml, version)
        self.parse_sites_response()

    def __iter__(self):
        for s in self.sites:
            yield s

    def __getitem__(self, key):
        if isinstance(key, int) and key < len(self.sites):
            return self.sites[key]

        if isinstance(key, str):
            site = [site for site in self.sites for code in site.site_info.site_codes if code == key]
            if len(site) > 0:
                return site[0]

        raise KeyError('Unknown key ' + str(key))

    def parse_sites_response(self, xml=None):
        """
        """
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        self.query_info = QueryInfo(self._find('queryInfo'), self._ns)
        self.sites = [Site(site, self._ns) for site in self._findall('site')]
        # except:
        #   raise ValueError('Cannot parse sitesResponse element correctly')

    """Accesability properties/methods"""
    @property
    def site_codes(self):
        return [site.site_info.site_codes for site in self.sites]

    @property
    def site_names(self):
        return [site.site_info.site_name for site in self.sites]


class QueryInfo(XMLParser):
    """
    """
    def __init__(self, xml_root, version='wml1.1'):
        super(QueryInfo, self).__init__(xml_root, version)
        self.parse_query_info()

    def parse_query_info(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
            # create queryinfo object from dict
        xml_dict = _xml_to_dict(self._root)
        self.creation_time = parser.parse(xml_dict.get('creation_time')) \
            if xml_dict.get('creation_time') is not None else None
        self.notes = [testXMLValue(note) for note in self._findall('note')]
        self.criteria = Criteria(self._find('criteria'), self._ns)
        # except:
        #   raise ValueError('Unable to parse queryInfo element correctly')


class Criteria(XMLParser):
    """
    """
    def __init__(self, xml_root, version='wml1.1'):
        super(Criteria, self).__init__(xml_root, version)
        self.parse_criteria()

    def parse_criteria(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        xml_dict = _xml_to_dict(self._root, depth=4)
        self.method_called = self._root.attrib.get('MethodCalled')
        self.location_param = xml_dict.get('location_param')
        self.variable_param = xml_dict.get('variable_param')
        try:
            self.begin_date_time = parser.parse(xml_dict['begin_date_time'])
        except Exception:
            self.begin_date_time = None

        try:
            self.end_date_time = parser.parse(xml_dict['end_date_time'])
        except Exception:
            self.end_date_time = None

        self.parameters = [(
            param.attrib.get('name'),
            param.attrib.get('value')) for param in self._findall('parameter')]
        # except:
        #   raise ValueError('Unable to parse xml for criteria element')


class Site(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Site, self).__init__(xml, version)
        self.parse_site()

    def __iter__(self):
        for c in self.series_catalogs:
            yield c

    def __getitem__(self, key):
        if isinstance(key, int) and key < len(self.series_catalogs):
            return self.series_catalogs[key]

        if isinstance(key, str):
            var = [series.variable for catalog in self.series_catalogs for series in catalog if series.code == key]
            if len(var) > 0:
                return var[0]

        raise KeyError('Unknown key ' + str(key))

    """Accessor propeties/methods"""
    @property
    def name(self):
        return self.site_info.site_name

    @property
    def codes(self):
        return self.site_info.site_codes

    @property
    def variable_names(self):
        return list(set([series.variable.variable_name for catalog in self.series_catalogs for series in catalog]))

    @property
    def variable_codes(self):
        return list(set([series.variable.variable_code for catalog in self.series_catalogs for series in catalog]))

    @property
    def geo_coords(self):
        return self.site_info.location.geo_coords

    @property
    def latitudes(self):
        return [g[1] for g in self.site_info.location.geo_coords]

    @property
    def longitudes(self):
        return [g[0] for g in self.site_info.location.geo_coords]

    def parse_site(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        self.site_info = SiteInfo(self._find('siteInfo'), self._ns)
        self.series_catalogs = [SeriesCatalog(elm, self._ns) for elm in self._findall('seriesCatalog')]
        # self.extension = Extension(self._find('extension'), self._ns)
        # except:
        #   raise ValueError('Unable to parse site element correctly')


class SiteInfo(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(SiteInfo, self).__init__(xml, version)
        self.parse_siteinfo()

    def parse_siteinfo(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        xml_dict = _xml_to_dict(self._root)
        self.site_name = xml_dict.get('site_name')
        self.site_codes = [testXMLValue(code) for code in self._findall('siteCode')]
        self.elevation = xml_dict.get('elevation_m')
        self.vertical_datum = xml_dict.get('vertical_datum')
        self.site_types = [testXMLValue(typ) for typ in self._findall('siteType')]
        self.site_properties = dict([
            (prop.attrib.get('name'), testXMLValue(prop)) for prop in self._findall('siteProperty')])
        self.altname = xml_dict.get('altname')
        self.notes = [testXMLValue(note) for note in self._findall('note')]
        # sub-objects
        tzi = self._find('timeZoneInfo')
        if tzi is not None:
            self.time_zone_info = TimeZoneInfo(tzi, self._ns)

        self.location = Location(self._find('geoLocation'), self._ns)

        # except:
        #   raise ValueError('Unable to parse siteInfo element')


class Location(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Location, self).__init__(xml, version)
        self.parse_location()

    def parse_location(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        xml_dict = _xml_to_dict(self._root)  # noqa: F841
        geogs = self._findall('geogLocation')
        self.geo_coords = list()
        self.srs = list()
        for g in geogs:
            self.geo_coords.append((
                testXMLValue(g.find(ns(self._ns) + 'longitude')),
                testXMLValue(g.find(ns(self._ns) + 'latitude'))))
            self.srs.append(g.attrib.get('srs'))

        locsite = self._findall('localSiteXY')
        self.local_sites = list()
        self.notes = list()
        self.projections = list()
        for ls in locsite:
            z = testXMLValue(ls.find(ns(self._ns) + 'Z'))
            if z is not None:
                self.local_sites.append((
                    testXMLValue(ls.find(ns(self._ns) + 'X')),
                    testXMLValue(ls.find(ns(self._ns) + 'Y')),
                    z))
            else:
                self.local_sites.append((
                    testXMLValue(ls.find(ns(self._ns) + 'X')),
                    testXMLValue(ls.find(ns(self._ns) + 'Y')),
                    '0'))

            self.notes.append([testXMLValue(note) for note in ls.findall(ns(self._ns) + 'note')])
            self.projections.append(ls.attrib.get('projectionInformation'))

        # except:
        #   raise ValueError('Unable to parse geoLocation element')


class TimeZoneInfo(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(TimeZoneInfo, self).__init__(xml, version)
        self.parse_timezoneinfo()

    def parse_timezoneinfo(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        xml_dict = _xml_to_dict(self._root)  # noqa: F841
        default = self._find('defaultTimeZone')
        if default is not None:
            self.zone_offset = default.attrib.get('zoneOffset')
            self.zone_abbreviation = default.attrib.get('zoneAbbreviation')

        daylight = self._find('daylightSavingsTimeZone')
        if daylight is not None:
            self.daylight_zone_offset = daylight.attrib.get('zoneOffset')
            self.daylight_zone_abbreviation = daylight.attrib.get('zoneAbbreviation')

        # except:
        #   raise ValueError('Unable to properly parset the timeZoneInfo element')


class SeriesCatalog(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(SeriesCatalog, self).__init__(xml, version)
        self.parse_seriescatalog()

    def __iter__(self):
        for s in self.series:
            yield s

    def __getitem__(self, key):
        if isinstance(key, int) and key < len(self.series):
            return self.series[key]

        if isinstance(key, str):
            srs = [series for series in self.series if series.code == key]
            if len(srs) > 0:
                return srs[0]

        raise KeyError('Unknown key ' + str(key))

    def parse_seriescatalog(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        self.series = [Series(elm, self._ns) for elm in self._findall('series')]
        # except:
        #   raise ValueError('Unable to properly parse the seriesCatalog element')


class Series(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Series, self).__init__(xml, version)
        self.parse_series()

    """Accessor proeprties/methods"""
    @property
    def name(self):
        return self.variable.variable_name

    @property
    def code(self):
        return self.variable.variable_code

    def parse_series(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        xml_dict = _xml_to_dict(self._root, depth=3)
        self.value_count = xml_dict.get('value_count')
        self.value_type = xml_dict.get('value_type')
        self.general_category = xml_dict.get('general_category')
        self.sample_medium = xml_dict.get('sample_medium')
        self.data_type = xml_dict.get('data_type')
        # date-time
        self.begin_date_time = parser.parse(xml_dict.get('begin_date_time'))
        self.begin_date_time_utc = parser.parse(xml_dict.get('begin_date_time_utc')) \
            if xml_dict.get('begin_date_time_utc') is not None else None
        self.end_date_time = parser.parse(xml_dict.get('end_date_time'))
        self.end_date_time_utc = parser.parse(xml_dict.get('end_date_time_utc')) \
            if xml_dict.get('end_date_time_utc') is not None else None
        # method info
        self.method_description = xml_dict.get('method_description')
        self.method_code = xml_dict.get('method_code')
        self.method_link = xml_dict.get('method_link')
        method = self._find('method')
        if method is not None:
            self.method_id = method.attrib.get('methodID')
        else:
            self.method_id = None

        # source info
        self.organization = xml_dict.get('organization')
        self.source_description = xml_dict.get('source_description')
        self.citation = xml_dict.get('citation')
        source = self._find('source')
        if source is not None:
            self.source_id = source.attrib.get('sourceID')
        else:
            self.source_id = None

        # quality control info
        self.quality_control_level_code = xml_dict.get('quality_control_level_code')
        self.definition = xml_dict.get('definition')
        qa = self._find('qualityControlLevel')
        if qa is not None:
            self.quality_control_level_id = qa.attrib.get('qualityControlLevelID')
        else:
            self.quality_control_level_id = None

        # properties
        self.properties = dict([(
            prop.attrib.get('name'),
            testXMLValue(prop)) for prop in self._findall('seriesProperty')])
        # sub-objects
        self.variable = Variable(self._find('variable'), self._ns)
        # except:
        #   raise ValueError('Unable to correctly parse Series element')


class Variable(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Variable, self).__init__(xml, version)
        self.parse_variable()

    def parse_variable(self, xml=None):
        if xml is not None:
            try:
                self._root = etree.parse(xml)
            except Exception:
                self._root = xml

        # try:
        xml_dict = _xml_to_dict(self._root)
        self.value_type = xml_dict.get('value_type')
        self.data_type = xml_dict.get('data_type')
        self.general_category = xml_dict.get('general_category')
        self.sample_medium = xml_dict.get('sample_medium')
        self.no_data_value = xml_dict.get('no_data_value')
        self.variable_name = xml_dict.get('variable_name')
        self.variable_code = xml_dict.get('variable_code')
        self.variable_description = xml_dict.get('variable_description')
        self.speciation = xml_dict.get('speciation')
        # notes and properties
        notes = [(note.attrib.get('title'), testXMLValue(note)) for note in self._findall('note')]
        none_notes = [note[1] for note in notes if note[0] is None]
        self.notes = dict([note for note in notes if note[0] is not None])
        if len(none_notes) > 0:
            self.notes['none'] = none_notes

        self.properties = dict([(
            prop.attrib.get('name'),
            testXMLValue(prop)) for prop in self._findall('variableProperty')])
        # related
        related = self._find('related')
        if related is not None:
            self.parent_codes = [dict([
                ('network', code.attrib.get('network')),
                ('vocabulary', code.attrib.get('vocabulary')),
                ('default', code.attrib.get('default'))]) for code in related.findall(ns(self._ns) + 'parentCode')]
            self.related_codes = [dict([
                ('network', code.get('network')),
                ('vocabulary', code.get('vocabulary')),
                ('default', code.get('default'))]) for code in related.findall(ns(self._ns) + 'relatedCode')]
        else:
            self.parent_codes = None
            self.related_codes = None

        # sub-objects
        if self._ns == 'wml1.0':
            unit = self._find('units')
            self.unit = Unit1_0(unit, self._ns) if unit is not None else None

            timesupport = self._find('timeSupport')
            self.time_support = TimeScale(timesupport, self._ns) if timesupport is not None else None
        else:
            unit = self._find('unit')
            self.unit = Unit(unit, self._ns) if unit is not None else None

            timescale = self._find('timeScale')
            self.time_scale = TimeScale(timescale, self._ns) if timescale is not None else None

        categories = self._find('categories')
        if categories is not None:
            self.categories = [Category(cat, self._ns) for cat in categories.findall(ns(self._ns) + 'category')]
        else:
            self.categories = None
        # except:
        #   raise ValueError('Unable to correctly parse variable element')


class TimeScale(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(TimeScale, self).__init__(xml, version)
        self.parse_timescale()

    def parse_timescale(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.time_spacing = xml_dict.get('time_spacing')
            self.time_support = xml_dict.get('time_support')
            self.time_interval = xml_dict.get('time_interval')
            unit = self._find('unit')
            self.unit = Unit(unit, self._ns) if unit is not None else None
        except Exception:
            raise


class Unit(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Unit, self).__init__(xml, version)
        self.parse_unit()

    def parse_unit(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.name = xml_dict.get('unit_name')
            self.unit_type = xml_dict.get('unit_type')
            self.description = xml_dict.get('unit_description')
            self.abbreviation = xml_dict.get('unit_abbreviation')
            self.code = xml_dict.get('unit_code')
            self.id = self._root.attrib.get('UnitID')
        except Exception:
            raise


class Unit1_0(XMLParser):
    def __init__(self, xml, version='wml1.0'):
        super(Unit1_0, self).__init__(xml, version)
        self.parse_unit()

    def parse_unit(self):
        try:
            self.name = testXMLValue(self._root)
            self.code = self._root.attrib.get('unitsCode')
            self.abbreviation = self._root.attrib.get('unitsAbbreviation')
            self.type = self._root.attrib.get('unitsType')
            self.id = self._root.attrib.get('unitID')
        except Exception:
            raise


class Category(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Category, self).__init__(xml, version)
        self.parse_category()

    def parse_category(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.data_value = xml_dict.get('data_value')
            self.description = xml_dict.get('description')
            self.id = self._root.attrib.get('categoryID')
        except Exception:
            raise


class TimeSeriesResponse(XMLParser):
    """
        Parses the response from a 'GetValues' request

        Parameters
        ===========
        :xmlio - A file-like object that holds the xml response from the request.

        Return
        =======
        An object constructed from a dictionary parse of the response. The object has get access and can
        also iterate over each timeSeries element returned.
    """
    def __init__(self, xml, version='wml1.1'):
        super(TimeSeriesResponse, self).__init__(xml, version)
        self.parse_timeseriesresponse()

    """Accessor properties/methods"""
    @property
    def series_names(self):
        return [series.name for series in self.time_series]

    @property
    def variable_names(self):
        return list(set([series.variable.variable_name for series in self.time_series]))

    @property
    def variable_codes(self):
        return list(set([s.variable.variable_code for s in self.time_series]))

    def get_series_by_variable(self, var_name=None, var_code=None):
        if var_code is not None:
            return [s for s in self.time_series if s.variable.variable_code == var_code]

        elif var_name is not None:
            return [series for series in self.time_series if series.variable.variable_name == var_name]

        return None

    def parse_timeseriesresponse(self):
        try:
            qi = self._find('queryInfo')
            self.query_info = QueryInfo(qi, self._ns)
            self.time_series = [TimeSeries(series, self._ns) for series in self._findall('timeSeries')]
        except Exception:
            raise


class TimeSeries(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(TimeSeries, self).__init__(xml, version)
        self.parse_timeseries()

    def parse_timeseries(self):
        try:
            self.variable = Variable(self._find('variable'), self._ns)
            self.values = [Values(val, self._ns) for val in self._findall('values')]
            self.source_info = SiteInfo(self._find('sourceInfo'), self._ns)
            self.name = self._root.attrib.get('name')
        except Exception:
            raise


class Values(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Values, self).__init__(xml, version)
        self.parse_values()

    def __iter__(self):
        for v in self.values:
            yield v

    """Accessor properties/methods"""
    def get_date_values(self, method_id=None, source_id=None, sample_id=None, quality_level=None, utc=False):
        varl = [v for v in self.values]
        if method_id is not None:
            varl = [v for v in varl if v.method_id == method_id]

        if source_id is not None:
            varl = [v for v in varl if v.source_id == source_id]

        if sample_id is not None:
            varl = [v for v in varl if v.sample_id == sample_id]

        if quality_level is not None:
            varl = [v for v in varl if v.quality_control_level == quality_level]

        if not utc:
            return [(v.date_time, v.value) for v in varl]
        else:
            return [(v.date_time_utc, v.value) for v in varl]

    def parse_values(self):
        xml_dict = _xml_to_dict(self._root)  # noqa: F841
        # method info
        self.methods = [Method(method, self._ns) for method in self._findall('method')]

        # source info
        self.sources = [Source(source, self._ns) for source in self._findall('source')]

        # quality control info
        self.qualit_control_levels = [
            QualityControlLevel(qal, self._ns) for qal in self._findall('qualityControlLevel')]

        # offset info
        self.offsets = [Offset(os, self._ns) for os in self._findall('offset')]

        # sample info
        self.samples = [Sample(sample, self._ns) for sample in self._findall('sample')]

        # censor codes
        self.censor_codes = [CensorCode(code, self._ns) for code in self._findall('censorCode')]

        # unit
        if self._ns == 'wml1.0':
            self.unit_abbreviation = self._root.attrib.get('unitsAbbreviation')
            self.unit_code = self._root.attrib.get('unitsCode')
            self.count = self._root.attrib.get('count')
        else:
            unit = self._find('unit')
            self.unit = Unit(unit, self._ns) if unit is not None else None

        # values
        self.values = [Value(val, self._ns) for val in self._findall('value')]


class Value(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Value, self).__init__(xml, version)
        self.parse_value()

    def parse_value(self):
        try:
            self.value = testXMLValue(self._root)
            d = self._root.attrib
            self.qualifiers = d.get('qualifiers')
            self.censor_code = d.get('censorCode')
            self.date_time = parser.parse(d.get('dateTime')) if d.get('dateTime') is not None else None
            self.time_offset = d.get('timeOffset')
            self.date_time_utc = parser.parse(d.get('dateTimeUTC')) if d.get('dateTimeUTC') is not None else None
            self.method_id = d.get('methodID')
            self.source_id = d.get('sourceID')
            self.accuracy_std_dev = d.get('accuracyStdDev')
            self.sample_id = d.get('sampleID')
            self.method_code = d.get('methodCode')
            self.source_code = d.get('sourceCode')
            self.lab_sample_code = d.get('lab_sample_code')
            self.offset_value = d.get('offsetValue')
            self.offset_type_id = d.get('offsetTypeID')
            self.offset_type_code = d.get('offsetTypeCode')
            self.coded_vocabulary = d.get('codedVocabulary')
            self.coded_vocabulary_term = d.get('codedVocabularyTerm')
            self.quality_control_level = d.get('qualityControlLevel')
            self.metadata_time = d.get('metadataTime')
            self.oid = d.get('oid')
        except Exception:
            raise


class Sample(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Sample, self).__init__(xml, version)
        self.parse_sample()

    def parse_sample(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.code = xml_dict.get('lab_sample_code')
            self.type = xml_dict.get('sample_type')
            lm = self._find('labMethod')
            self.method = LabMethod(lm, self._ns) if lm is not None else None
        except Exception:
            raise


class LabMethod(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(LabMethod, self).__init__(xml, version)
        self.parse_labmethod()

    def parse_labmethod(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.code = xml_dict.get('lab_code')
            self.name = xml_dict.get('lab_name')
            self.organization = xml_dict.get('lab_organization')
            self.method_name = xml_dict.get('lab_method_name')
            self.method_description = xml_dict.get('lab_method_description')
            self.method_link = xml_dict.get('lab_method_link')
            # sub-objects
            source = self._find('labSourceDetails')
            self.source_details = Source(source, self._ns) if source is not None else None
        except Exception:
            raise


class Source(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Source, self).__init__(xml, version)
        self.parse_source()

    def __str__(self):
        return str(self.__dict__)

    def get_contact(self, name):
        ci = [ci for ci in self.contact_info if ci.name == name]
        if len(ci) < 0:
            return ci[0]
        return None

    def parse_source(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.code = xml_dict.get('source_code')
            self.organization = xml_dict.get('organization')
            self.description = xml_dict.get('source_description')
            self.links = [testXMLValue(link) for link in self._findall('sourceLink')]
            self.citation = xml_dict.get('citation')
            # metadata
            self.topic_category = xml_dict.get('topic_category')
            self.title = xml_dict.get('title')
            self.abstract = xml_dict.get('abstract')
            self.profile_version = xml_dict.get('profile_version')
            self.metadata_link = xml_dict.get('metadata_link')
            # contact info
            self.contact_info = [ContactInformation(ci, self._ns) for ci in self._findall('contactInformation')]
        except Exception:
            raise


class ContactInformation(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(ContactInformation, self).__init__(xml, version)
        self.parse_contactinformation()

    def parse_contactinformation(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.name = xml_dict.get('contact_name')
            self.type = xml_dict.get('type_of_contact')
            self.email = [testXMLValue(email) for email in self._findall('email')]
            self.phone = [testXMLValue(phone) for phone in self._findall('phone')]
            self.address = [testXMLValue(address) for address in self._findall('address')]
        except Exception:
            raise


class Offset(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Offset, self).__init__(xml, version)
        self.parse_offset()

    def parse_offset(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.type_code = xml_dict.get('offset_type_code')
            self.value = xml_dict.get('offset_value')
            self.description = xml_dict.get('offset_description')
            self.is_vertical = xml_dict.get('offset_is_vertical')
            self.azimuth_degrees = xml_dict.get('offset_azimuth_degrees')
            unit = self._root.find('unit')
            if self._ns == 'wml1.0':
                self.unit = Unit1_0(unit, self._ns) if unit is not None else None
            else:
                self.unit = Unit(unit, self._ns) if unit is not None else None
        except Exception:
            raise


class Method(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(Method, self).__init__(xml, version)
        self.parse_method()

    def parse_method(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.code = xml_dict.get('method_code')
            self.description = xml_dict.get('method_description')
            self.link = xml_dict.get('method_link')
            self.id = self._root.attrib.get('methodID')
        except Exception:
            raise


class QualityControlLevel(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(QualityControlLevel, self).__init__(xml, version)
        self.parse_qcl()

    def parse_qcl(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.code = xml_dict.get('quality_control_level_code')
            self.definition = xml_dict.get('definition')
            self.explanation = xml_dict.get('explanation')
            self.id = self._root.attrib.get('qualityControlLevelID')
        except Exception:
            raise


class CensorCode(XMLParser):
    def __init__(self, xml, version='wml1.1'):
        super(CensorCode, self).__init__(xml, version)
        self.parse_censorcode()

    def parse_censorcode(self):
        try:
            xml_dict = _xml_to_dict(self._root)
            self.code = xml_dict.get('censor_code')
            self.description = xml_dict.get('censor_code_description')
            self.id = self._root.attrib.get('censorCodeID')
        except Exception:
            raise


class VariablesResponse(XMLParser):
    """
        Parses the response from a 'GetVariableInfo' request

        Parameters
        ===========
        :xmlio - A file-like object that holds the xml response from the request.

        Return
        =======
        An object constructed from a dictionary parse of the response. The object has get access to its variables and
        can also be used as an iterator.
    """
    def __init__(self, xml, version='wml1.1'):
        super(VariablesResponse, self).__init__(xml, version)
        self.parse_variablesresponse()

    def __iter__(self):
        for v in self.variables:
            yield v

    def __getitem__(self, key):
        if isinstance(key, int) and key < len(self.variables):
            return self.variables[key]

        if isinstance(key, str):
            v = [var for var in self.variables if var.variable_code == key]
            if len(v) > 0:
                return v[0]

            v = [var for var in self.variables if var.variable_name == key]
            if len(v) > 0:
                return v[0]

        raise KeyError('Unknown key ' + str(key))

    """Accessor properties/methods"""
    @property
    def variable_names(self):
        return list(set([var.variable_name for var in self.variables]))

    @property
    def variable_codes(self):
        return [var.variable_code for var in self.variables]

    def parse_variablesresponse(self):
        try:
            qi = self._find('queryInfo')
            self.query_info = QueryInfo(qi, self._ns) if qi is not None else None
            varis = self._find('variables')
            self.variables = [Variable(var, self._ns) for var in varis.findall(ns(self._ns) + 'variable')]
        except Exception:
            raise



# =============================================================================
# Copyright (c) 2020 Tom Kralidis
#
# Author: Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

import json
import logging

from urllib.parse import urljoin

from owslib.ogcapi import API, REQUEST_HEADERS
from owslib.util import http_get

LOGGER = logging.getLogger(__name__)


class Features(API):
    """Abstraction for OGC API - Features"""

    def __init__(self, url, json_=None, timeout=30, headers=None, auth=None):
        __doc__ = API.__doc__  # noqa
        super().__init__(url, json_, timeout, headers, auth)

    def collection_items(self, collection_id, **kwargs):
        """
        implements /collection/{collectionId}/items

        @type collection_id: string
        @param collection_id: id of collection
        @type bbox: list
        @param bbox: list of minx,miny,maxx,maxy
        @type datetime: string
        @param datetime: time extent or time instant
        @type limit: int
        @param limit: limit number of features
        @type startindex: int
        @param startindex: start position of results

        @returns: feature results
        """

        if 'bbox' in kwargs:
            kwargs['bbox'] = ','.join(kwargs['bbox'])

        path = 'collections/{}/items'.format(collection_id)
        url = self._build_url(path)
        LOGGER.debug('Request: {}'.format(url))
        response = http_get(
            url, headers=self.headers, params=kwargs, auth=self.auth
        ).json()
        return response

    def collection_item(self, collection_id, identifier):
        """
        implements /collections/{collectionId}/items/{featureId}

        @type collection_id: string
        @param collection_id: id of collection
        @type identifier: string
        @param identifier: feature identifier

        @returns: single feature result
        """

        path = 'collections/{}/items/{}'.format(collection_id, identifier)
        url = self._build_url(path)
        LOGGER.debug('Request: {}'.format(url))
        response = http_get(url, headers=self.headers, auth=self.auth).json()
        return response



from urllib.parse import urlencode, parse_qsl

from owslib.etree import etree
from owslib.util import strip_bom, Authentication, openURL


class WMSCapabilitiesReader(object):
    """Read and parse capabilities document into a lxml.etree infoset
    """

    def __init__(self, version='1.1.1', url=None, un=None, pw=None, headers=None, auth=None):
        """Initialize"""
        self.version = version
        self._infoset = None
        self.url = url
        if auth:
            if un:
                auth.username = un
            if pw:
                auth.password = pw
        self.headers = headers
        self.request = None
        self.auth = auth or Authentication(un, pw)

        # if self.username and self.password:
        #     # Provide login information in order to use the WMS server
        #     # Create an OpenerDirector with support for Basic HTTP
        #     # Authentication...
        #     passman = HTTPPasswordMgrWithDefaultRealm()
        #     passman.add_password(None, self.url, self.username, self.password)
        #     auth_handler = HTTPBasicAuthHandler(passman)
        #     opener = build_opener(auth_handler)
        #     self._open = opener.open

    def capabilities_url(self, service_url):
        """Return a capabilities url
        """
        qs = []
        if service_url.find('?') != -1:
            qs = parse_qsl(service_url.split('?')[1])

        params = [x[0] for x in qs]

        if 'service' not in params:
            qs.append(('service', 'WMS'))
        if 'request' not in params:
            qs.append(('request', 'GetCapabilities'))
        if 'version' not in params:
            qs.append(('version', self.version))

        urlqs = urlencode(tuple(qs))
        return service_url.split('?')[0] + '?' + urlqs

    def read(self, service_url, timeout=30):
        """Get and parse a WMS capabilities document, returning an
        elementtree instance

        service_url is the base url, to which is appended the service,
        version, and request parameters
        """
        self.request = self.capabilities_url(service_url)

        # now split it up again to use the generic openURL function...
        spliturl = self.request.split('?')
        u = openURL(spliturl[0], spliturl[1], method='Get',
                    timeout=timeout, headers=self.headers, auth=self.auth)

        raw_text = strip_bom(u.read())
        return etree.fromstring(raw_text)

    def readString(self, st):
        """Parse a WMS capabilities document, returning an elementtree instance.

        string should be an XML capabilities document
        """
        if not isinstance(st, str) and not isinstance(st, bytes):
            raise ValueError("String must be of type string or bytes, not %s" % type(st))
        raw_text = strip_bom(st)
        return etree.fromstring(raw_text)


class AbstractContentMetadata(object):

    def __init__(self, auth=None):
        self.auth = auth or Authentication()

    def get_metadata(self):
        return [m['metadata'] for m in self.metadataUrls if m.get('metadata', None) is not None]



# =============================================================================
# OWSLib. Copyright (C) 2015 Jachym Cepicky
#
# Contact email: jachym.cepicky@gmail.com
#
# =============================================================================
"""
Set of functions, which are suitable for DescribeFeatureType parsing and
generating layer schema description compatible with `fiona`
"""

import sys
from urllib.parse import urlencode, parse_qsl
from owslib.etree import etree
from owslib.namespaces import Namespaces
from owslib.util import which_etree, findall, Authentication, openURL

MYNS = Namespaces()
XS_NAMESPACE = MYNS.get_namespace("xs")
GML_NAMESPACES = (
    MYNS.get_namespace("gml"),
    MYNS.get_namespace("gml311"),
    MYNS.get_namespace("gml32"),
)


def get_schema(
    url, typename, version="1.0.0", timeout=30, headers=None, username=None, password=None, auth=None
):
    """Parses DescribeFeatureType response and creates schema compatible
    with :class:`fiona`

    :param str url: url of the service
    :param str version: version of the service
    :param str typename: name of the layer
    :param int timeout: request timeout
    :param str username: service authentication username
    :param str password: service authentication password
    :param Authentication auth: instance of owslib.util.Authentication
    """
    if auth:
        if username:
            auth.username = username
        if password:
            auth.password = password
    else:
        auth = Authentication(username, password)
    url = _get_describefeaturetype_url(url, version, typename)
    root = _get_remote_describefeaturetype(url, timeout=timeout,
                                           headers=headers, auth=auth)

    if ":" in typename:
        typename = typename.split(":")[1]
    type_element = root.find("./{%s}element" % XS_NAMESPACE)
    complex_type = type_element.attrib["type"].split(":")[1]
    elements = _get_elements(complex_type, root)
    nsmap = None
    if hasattr(root, "nsmap"):
        nsmap = root.nsmap
    return _construct_schema(elements, nsmap)


def _get_elements(complex_type, root):
    """Get attribute elements
    """

    found_elements = []
    element = findall(
        root,
        "{%s}complexType" % XS_NAMESPACE,
        attribute_name="name",
        attribute_value=complex_type,
    )[0]
    found_elements = findall(element, "{%s}element" % XS_NAMESPACE)

    return found_elements


def _construct_schema(elements, nsmap):
    """Consruct fiona schema based on given elements

    :param list Element: list of elements
    :param dict nsmap: namespace map

    :return dict: schema
    """

    schema = {"properties": {}, "required": [], "geometry": None}

    schema_key = None
    gml_key = None

    # if nsmap is defined, use it
    if nsmap:
        for key in nsmap:
            if nsmap[key] == XS_NAMESPACE:
                schema_key = key
            if nsmap[key] in GML_NAMESPACES:
                gml_key = key
    # if no nsmap is defined, we have to guess
    else:
        gml_key = "gml"
        schema_key = "xsd"

    mappings = {
        "PointPropertyType": "Point",
        "PolygonPropertyType": "Polygon",
        "LineStringPropertyType": "LineString",
        "MultiPointPropertyType": "MultiPoint",
        "MultiLineStringPropertyType": "MultiLineString",
        "MultiPolygonPropertyType": "MultiPolygon",
        "MultiGeometryPropertyType": "MultiGeometry",
        "GeometryPropertyType": "GeometryCollection",
        "SurfacePropertyType": "3D Polygon",
        "MultiSurfacePropertyType": "3D MultiPolygon",
    }

    for element in elements:
        data_type = element.attrib["type"].replace(gml_key + ":", "")
        name = element.attrib["name"]
        non_nillable = element.attrib.get("nillable", "false") == "false"

        if data_type in mappings:
            schema["geometry"] = mappings[data_type]
            schema["geometry_column"] = name
        else:
            schema["properties"][name] = data_type.replace(schema_key + ":", "")

        if non_nillable:
            schema["required"].append(name)

    if schema["properties"] or schema["geometry"]:
        return schema
    else:
        return None


def _get_describefeaturetype_url(url, version, typename):
    """Get url for describefeaturetype request

    :return str: url
    """

    query_string = []
    if url.find("?") != -1:
        query_string = parse_qsl(url.split("?")[1])

    params = [x[0] for x in query_string]

    if "service" not in params:
        query_string.append(("service", "WFS"))
    if "request" not in params:
        query_string.append(("request", "DescribeFeatureType"))
    if "version" not in params:
        query_string.append(("version", version))

    query_string.append(("typeName", typename))

    urlqs = urlencode(tuple(query_string))
    return url.split("?")[0] + "?" + urlqs


def _get_remote_describefeaturetype(url, timeout, headers, auth):
    """Gets the DescribeFeatureType response from the remote server.

    :param str url: url of the service
    :param int timeout: request timeout
    :param Authentication auth: instance of owslib.util.Authentication

    :return etree.Element with the root of the DescribeFeatureType response
    """
    res = openURL(url, timeout=timeout, headers=headers, auth=auth)
    return etree.fromstring(res.read())



# -*- coding: iso-8859-15 -*-
# =============================================================================
# Copyright (c) 2004, 2006 Sean C. Gillies
# Copyright (c) 2005 Nuxeo SARL <http://nuxeo.com>
#
# Authors : Sean Gillies <sgillies@frii.com>
#           Julien Anguenot <ja@nuxeo.com>
#
# Contact email: sgillies@frii.com
# =============================================================================

"""
API for Web Map Service (WMS) methods and metadata.

Currently supports only version 1.1.1 of the WMS protocol.
"""

import cgi
from urllib.parse import urlencode

import warnings

from owslib.etree import etree
from owslib.util import (openURL, testXMLValue, extract_xml_list,
                         xmltag_split, OrderedDict, ServiceException,
                         bind_url, nspath_eval, Authentication)
from owslib.fgdc import Metadata
from owslib.iso import MD_Metadata
from owslib.map.common import WMSCapabilitiesReader, AbstractContentMetadata
from owslib.namespaces import Namespaces

n = Namespaces()


class CapabilitiesError(Exception):
    pass


class WebMapService_1_1_1(object):
    """Abstraction for OGC Web Map Service (WMS).

    Implements IWebMapService.
    """

    def __getitem__(self, name):
        ''' check contents dictionary to allow dict
        like access to service layers
        '''
        if name in self.__getattribute__('contents'):
            return self.__getattribute__('contents')[name]
        else:
            raise KeyError("No content named %s" % name)

    def __init__(self, url, version='1.1.1', xml=None, username=None, password=None,
                 parse_remote_metadata=False, headers=None, timeout=30, auth=None):
        """Initialize."""
        if auth:
            if username:
                auth.username = username
            if password:
                auth.password = password
        self.url = url
        self.version = version
        self.timeout = timeout
        self.headers = headers
        self._capabilities = None
        self.auth = auth or Authentication(username, password)

        # Authentication handled by Reader
        reader = WMSCapabilitiesReader(
            self.version, url=self.url, headers=headers, auth=self.auth)
        if xml:  # read from stored xml
            self._capabilities = reader.readString(xml)
        else:  # read from server
            self._capabilities = reader.read(self.url, timeout=self.timeout)

        self.request = reader.request

        # avoid building capabilities metadata if the
        # response is a ServiceExceptionReport
        se = self._capabilities.find('ServiceException')
        if se is not None:
            err_message = str(se.text).strip()
            raise ServiceException(err_message)

        # build metadata objects
        self._buildMetadata(parse_remote_metadata)

    def _buildMetadata(self, parse_remote_metadata=False):
        """Set up capabilities metadata objects."""

        self.updateSequence = self._capabilities.attrib.get('updateSequence')

        # serviceIdentification metadata
        serviceelem = self._capabilities.find('Service')
        self.identification = ServiceIdentification(serviceelem, self.version)

        # serviceProvider metadata
        self.provider = ServiceProvider(serviceelem)

        # serviceOperations metadata
        self.operations = []
        for elem in self._capabilities.find('Capability/Request')[:]:
            self.operations.append(OperationMetadata(elem))

        # serviceContents metadata: our assumption is that services use a
        # top-level layer as a metadata organizer, nothing more.
        self.contents = OrderedDict()
        caps = self._capabilities.find('Capability')

        # recursively gather content metadata for all layer elements.
        # To the WebMapService.contents store only metadata of named layers.
        def gather_layers(parent_elem, parent_metadata):
            layers = []
            for index, elem in enumerate(parent_elem.findall('Layer')):
                cm = ContentMetadata(elem, parent=parent_metadata,
                                     index=index + 1,
                                     parse_remote_metadata=parse_remote_metadata)
                if cm.id:
                    if cm.id in self.contents:
                        warnings.warn('Content metadata for layer "%s" already exists. Using child layer' % cm.id)
                    layers.append(cm)
                    self.contents[cm.id] = cm
                cm.children = gather_layers(elem, cm)
            return layers
        gather_layers(caps, None)

        # exceptions
        self.exceptions = [f.text for f
                           in self._capabilities.findall('Capability/Exception/Format')]

    def items(self):
        '''supports dict-like items() access'''
        items = []
        for item in self.contents:
            items.append((item, self.contents[item]))
        return items

    def getcapabilities(self):
        """Request and return capabilities document from the WMS as a
        file-like object.
        NOTE: this is effectively redundant now"""

        reader = WMSCapabilitiesReader(
            self.version, url=self.url, auth=self.auth)
        u = self._open(reader.capabilities_url(self.url))
        # check for service exceptions, and return
        if u.info()['Content-Type'] == 'application/vnd.ogc.se_xml':
            se_xml = u.read()
            se_tree = etree.fromstring(se_xml)
            err_message = str(se_tree.find('ServiceException').text).strip()
            raise ServiceException(err_message)
        return u

    def __build_getmap_request(self, layers=None, styles=None, srs=None, bbox=None,
                               format=None, size=None, time=None, transparent=False,
                               bgcolor=None, exceptions=None, **kwargs):

        request = {'service': 'WMS', 'version': self.version, 'request': 'GetMap'}

        # check layers and styles
        assert len(layers) > 0
        request['layers'] = ','.join(layers)
        if styles:
            assert len(styles) == len(layers)
            request['styles'] = ','.join(styles)
        else:
            request['styles'] = ''

        # size
        request['width'] = str(size[0])
        request['height'] = str(size[1])

        request['srs'] = str(srs)
        request['bbox'] = ','.join([repr(x) for x in bbox])
        request['format'] = str(format)
        request['transparent'] = str(transparent).upper()
        request['bgcolor'] = '0x' + bgcolor[1:7]
        request['exceptions'] = str(exceptions)

        if time is not None:
            request['time'] = str(time)

        if kwargs:
            for kw in kwargs:
                request[kw] = kwargs[kw]
        return request

    def getmap(self, layers=None, styles=None, srs=None, bbox=None,
               format=None, size=None, time=None, transparent=False,
               bgcolor='#FFFFFF',
               exceptions='application/vnd.ogc.se_xml',
               method='Get',
               timeout=None,
               **kwargs
               ):
        """Request and return an image from the WMS as a file-like object.

        Parameters
        ----------
        layers : list
            List of content layer names.
        styles : list
            Optional list of named styles, must be the same length as the
            layers list.
        srs : string
            A spatial reference system identifier.
        bbox : tuple
            (left, bottom, right, top) in srs units.
        format : string
            Output image format such as 'image/jpeg'.
        size : tuple
            (width, height) in pixels.
        transparent : bool
            Optional. Transparent background if True.
        bgcolor : string
            Optional. Image background color.
        method : string
            Optional. HTTP DCP method name: Get or Post.
        **kwargs : extra arguments
            anything else e.g. vendor specific parameters

        Example
        -------
            wms = WebMapService('http://giswebservices.massgis.state.ma.us/geoserver/wms', version='1.1.1')
            img = wms.getmap(layers=['massgis:GISDATA.SHORELINES_ARC'],\
                                 styles=[''],\
                                 srs='EPSG:4326',\
                                 bbox=(-70.8, 42, -70, 42.8),\
                                 size=(300, 300),\
                                 format='image/jpeg',\
                                 transparent=True)
            out = open('example.jpg', 'wb')
            bytes_written = out.write(img.read())
            out.close()

        """
        try:
            base_url = next((m.get('url') for m in self.getOperationByName('GetMap').methods
                            if m.get('type').lower() == method.lower()))
        except StopIteration:
            base_url = self.url
        request = {'version': self.version, 'request': 'GetMap'}

        request = self.__build_getmap_request(
            layers=layers,
            styles=styles,
            srs=srs,
            bbox=bbox,
            format=format,
            size=size,
            time=time,
            transparent=transparent,
            bgcolor=bgcolor,
            exceptions=exceptions,
            **kwargs)

        data = urlencode(request)

        self.request = bind_url(base_url) + data

        u = openURL(base_url, data, method, timeout=timeout or self.timeout, auth=self.auth)

        # check for service exceptions, and return
        if u.info().get('Content-Type', '').split(';')[0] in ['application/vnd.ogc.se_xml']:
            se_xml = u.read()
            se_tree = etree.fromstring(se_xml)
            err_message = str(se_tree.find('ServiceException').text).strip()
            raise ServiceException(err_message)
        return u

    def getfeatureinfo(self,
                       layers=None,
                       styles=None,
                       srs=None,
                       bbox=None,
                       format=None,
                       size=None,
                       time=None,
                       transparent=False,
                       bgcolor='#FFFFFF',
                       exceptions='application/vnd.ogc.se_xml',
                       query_layers=None,
                       xy=None,
                       info_format=None,
                       feature_count=20,
                       method='Get',
                       timeout=None,
                       **kwargs
                       ):
        try:
            base_url = next((m.get('url') for m in self.getOperationByName('GetFeatureInfo').methods
                            if m.get('type').lower() == method.lower()))
        except StopIteration:
            base_url = self.url

        # GetMap-Request
        request = self.__build_getmap_request(
            layers=layers,
            styles=styles,
            srs=srs,
            bbox=bbox,
            format=format,
            size=size,
            time=time,
            transparent=transparent,
            bgcolor=bgcolor,
            exceptions=exceptions,
            **kwargs)

        # extend to GetFeatureInfo-Request
        request['request'] = 'GetFeatureInfo'

        if not query_layers:
            __str_query_layers = ','.join(layers)
        else:
            __str_query_layers = ','.join(query_layers)

        request['query_layers'] = __str_query_layers
        request['x'] = str(xy[0])
        request['y'] = str(xy[1])
        request['info_format'] = info_format
        request['feature_count'] = str(feature_count)

        data = urlencode(request)

        self.request = bind_url(base_url) + data

        u = openURL(base_url, data, method, timeout=timeout or self.timeout, auth=self.auth)

        # check for service exceptions, and return
        if u.info()['Content-Type'] == 'application/vnd.ogc.se_xml':
            se_xml = u.read()
            se_tree = etree.fromstring(se_xml)
            err_message = str(se_tree.find('ServiceException').text).strip()
            raise ServiceException(err_message)
        return u

    def getServiceXML(self):
        xml = None
        if self._capabilities is not None:
            xml = etree.tostring(self._capabilities)
        return xml

    def getOperationByName(self, name):
        """Return a named content item."""
        for item in self.operations:
            if item.name == name:
                return item
        raise KeyError("No operation named %s" % name)


class ServiceIdentification(object):
    ''' Implements IServiceIdentificationMetadata '''

    def __init__(self, infoset, version):
        self._root = infoset
        self.type = testXMLValue(self._root.find('Name'))
        self.version = version
        self.title = testXMLValue(self._root.find('Title'))
        self.abstract = testXMLValue(self._root.find('Abstract'))
        self.keywords = extract_xml_list(self._root.findall('KeywordList/Keyword'))
        self.accessconstraints = testXMLValue(self._root.find('AccessConstraints'))
        self.fees = testXMLValue(self._root.find('Fees'))


class ServiceProvider(object):
    ''' Implements IServiceProviderMetatdata '''
    def __init__(self, infoset):
        self._root = infoset
        name = self._root.find('ContactInformation/ContactPersonPrimary/ContactOrganization')
        if name is not None:
            self.name = name.text
        else:
            self.name = None
        self.url = None
        online_resource = self._root.find('OnlineResource')
        if online_resource is not None:
            self.url = online_resource.attrib.get('{http://www.w3.org/1999/xlink}href', '')
        # contact metadata
        contact = self._root.find('ContactInformation')
        # sometimes there is a contact block that is empty, so make
        # sure there are children to parse
        if contact is not None and contact[:] != []:
            self.contact = ContactMetadata(contact)
        else:
            self.contact = None

    def getContentByName(self, name):
        """Return a named content item."""
        for item in self.contents:
            if item.name == name:
                return item
        raise KeyError("No content named %s" % name)

    def getOperationByName(self, name):
        """Return a named content item."""
        for item in self.operations:
            if item.name == name:
                return item
        raise KeyError("No operation named %s" % name)


class ContentMetadata(AbstractContentMetadata):
    """
    Abstraction for WMS layer metadata.

    Implements IContentMetadata.
    """

    def __init__(self, elem, parent=None, children=None, index=0,
                 parse_remote_metadata=False, timeout=30, auth=None):
        super(ContentMetadata, self).__init__(auth)
        if elem.tag != 'Layer':
            raise ValueError('%s should be a Layer' % (elem,))

        self.parent = parent
        if parent:
            self.index = "%s.%d" % (parent.index, index)
        else:
            self.index = str(index)

        self._children = children

        self.id = self.name = testXMLValue(elem.find('Name'))

        # layer attributes
        self.queryable = int(elem.attrib.get('queryable', 0))
        self.cascaded = int(elem.attrib.get('cascaded', 0))
        self.opaque = int(elem.attrib.get('opaque', 0))
        self.noSubsets = int(elem.attrib.get('noSubsets', 0))
        self.fixedWidth = int(elem.attrib.get('fixedWidth', 0))
        self.fixedHeight = int(elem.attrib.get('fixedHeight', 0))

        # title is mandatory property
        self.title = None
        title = testXMLValue(elem.find('Title'))
        if title is not None:
            self.title = title.strip()

        self.abstract = testXMLValue(elem.find('Abstract'))

        # bboxes
        b = elem.find('BoundingBox')
        self.boundingBox = None
        if b is not None:
            try:  # sometimes the SRS attribute is (wrongly) not provided
                srs = b.attrib['SRS']
            except KeyError:
                srs = None
            self.boundingBox = (
                float(b.attrib['minx']),
                float(b.attrib['miny']),
                float(b.attrib['maxx']),
                float(b.attrib['maxy']),
                srs,
            )
        elif self.parent:
            if hasattr(self.parent, 'boundingBox'):
                self.boundingBox = self.parent.boundingBox

        # ScaleHint
        sh = elem.find('ScaleHint')
        self.scaleHint = None
        if sh is not None:
            if 'min' in sh.attrib and 'max' in sh.attrib:
                self.scaleHint = {'min': sh.attrib['min'], 'max': sh.attrib['max']}

        attribution = elem.find('Attribution')
        if attribution is not None:
            self.attribution = dict()
            title = attribution.find('Title')
            url = attribution.find('OnlineResource')
            logo = attribution.find('LogoURL')
            if title is not None:
                self.attribution['title'] = title.text
            if url is not None:
                self.attribution['url'] = url.attrib['{http://www.w3.org/1999/xlink}href']
            if logo is not None:
                self.attribution['logo_size'] = (int(logo.attrib['width']), int(logo.attrib['height']))
                self.attribution['logo_url'] = logo.find('OnlineResource').attrib['{http://www.w3.org/1999/xlink}href']

        b = elem.find('LatLonBoundingBox')
        if b is not None:
            self.boundingBoxWGS84 = (
                float(b.attrib['minx']),
                float(b.attrib['miny']),
                float(b.attrib['maxx']),
                float(b.attrib['maxy']),
            )
        elif self.parent:
            self.boundingBoxWGS84 = self.parent.boundingBoxWGS84
        else:
            self.boundingBoxWGS84 = None

        # SRS options
        self.crsOptions = []

        # Copy any parent SRS options (they are inheritable properties)
        if self.parent:
            self.crsOptions = list(self.parent.crsOptions)

        # Look for SRS option attached to this layer
        if elem.find('SRS') is not None:
            # some servers found in the wild use a single SRS
            # tag containing a whitespace separated list of SRIDs
            # instead of several SRS tags. hence the inner loop
            for srslist in [x.text for x in elem.findall('SRS')]:
                if srslist:
                    for srs in srslist.split():
                        self.crsOptions.append(srs)

        # Get rid of duplicate entries
        self.crsOptions = list(set(self.crsOptions))

        # Set self.crsOptions to None if the layer (and parents) had no SRS options
        if len(self.crsOptions) == 0:
            # raise ValueError('%s no SRS available!?' % (elem,))
            # Comment by D Lowe.
            # Do not raise ValueError as it is possible that a layer is purely a parent layer and does
            # not have SRS specified. Instead set crsOptions to None
            # Comment by Jachym:
            # Do not set it to None, but to [], which will make the code
            # work further. Fixed by anthonybaxter
            self.crsOptions = []

        # Styles
        self.styles = {}

        # Copy any parent styles (they are inheritable properties)
        if self.parent:
            self.styles = self.parent.styles.copy()

        # Get the styles for this layer (items with the same name are replaced)
        for s in elem.findall('Style'):
            name = s.find('Name')
            title = s.find('Title')
            if name is None or title is None:
                raise ValueError('%s missing name or title' % (s,))
            style = {'title': title.text}
            # legend url
            legend = s.find('LegendURL/OnlineResource')
            if legend is not None:
                style['legend'] = legend.attrib['{http://www.w3.org/1999/xlink}href']
            self.styles[name.text] = style

        # keywords
        self.keywords = [f.text for f in elem.findall('KeywordList/Keyword')]

        # timepositions - times for which data is available.
        self.timepositions = None
        self.defaulttimeposition = None
        for extent in elem.findall('Extent'):
            if extent.attrib.get("name").lower() == 'time':
                if extent.text:
                    self.timepositions = extent.text.split(',')
                    self.defaulttimeposition = extent.attrib.get("default")
                    break

        # Elevations - available vertical levels
        self.elevations = None
        for extent in elem.findall('Extent'):
            if extent.attrib.get("name").lower() == 'elevation':
                if extent.text:
                    self.elevations = extent.text.split(',')
                    break

        # MetadataURLs
        self.metadataUrls = []
        for m in elem.findall('MetadataURL'):
            metadataUrl = {
                'type': testXMLValue(m.attrib['type'], attrib=True),
                'format': testXMLValue(m.find('Format')),
                'url': testXMLValue(m.find('OnlineResource').attrib['{http://www.w3.org/1999/xlink}href'], attrib=True)
            }
            self.metadataUrls.append(metadataUrl)

        # DataURLs
        self.dataUrls = []
        for m in elem.findall('DataURL'):
            dataUrl = {
                'format': m.find('Format').text.strip(),
                'url': m.find('OnlineResource').attrib['{http://www.w3.org/1999/xlink}href']
            }
            self.dataUrls.append(dataUrl)

        self.layers = []
        for child in elem.findall('Layer'):
            self.layers.append(ContentMetadata(child, self))

    def parse_remote_metadata(self, timeout=30):
        """Parse remote metadata for MetadataURL and add it as metadataUrl['metadata']"""
        for metadataUrl in self.metadataUrls:
            if metadataUrl['url'] is not None \
                    and metadataUrl['format'].lower() in ['application/xml', 'text/xml']:  # download URL
                try:
                    content = openURL(
                        metadataUrl['url'], timeout=timeout, auth=self.auth)
                    doc = etree.fromstring(content.read())

                    if metadataUrl['type'] == 'FGDC':
                        mdelem = doc.find('.//metadata')
                        if mdelem is not None:
                            metadataUrl['metadata'] = Metadata(mdelem)
                            continue

                    if metadataUrl['type'] == 'TC211':
                        mdelem = doc.find('.//' + nspath_eval('gmd:MD_Metadata', n.get_namespaces(['gmd']))) \
                            or doc.find('.//' + nspath_eval('gmi:MI_Metadata', n.get_namespaces(['gmi'])))
                        if mdelem is not None:
                            metadataUrl['metadata'] = MD_Metadata(mdelem)
                            continue
                except Exception:
                    metadataUrl['metadata'] = None

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, value):
        if self._children is None:
            self._children = value
        else:
            self._children.extend(value)

    def __str__(self):
        return 'Layer Name: %s Title: %s' % (self.name, self.title)


class OperationMetadata:
    """Abstraction for WMS OperationMetadata.

    Implements IOperationMetadata.
    """
    def __init__(self, elem):
        """."""
        self.name = xmltag_split(elem.tag)
        # formatOptions
        self.formatOptions = [f.text for f in elem.findall('Format')]
        self.methods = []
        for verb in elem.findall('DCPType/HTTP/*'):
            url = verb.find('OnlineResource').attrib['{http://www.w3.org/1999/xlink}href']
            self.methods.append({'type': xmltag_split(verb.tag), 'url': url})


class ContactMetadata:
    """Abstraction for contact details advertised in GetCapabilities.
    """
    def __init__(self, elem):
        name = elem.find('ContactPersonPrimary/ContactPerson')
        if name is not None:
            self.name = name.text
        else:
            self.name = None
        email = elem.find('ContactElectronicMailAddress')
        if email is not None:
            self.email = email.text
        else:
            self.email = None
        self.address = self.city = self.region = None
        self.postcode = self.country = None

        address = elem.find('ContactAddress')
        if address is not None:
            street = address.find('Address')
            if street is not None:
                self.address = street.text

            city = address.find('City')
            if city is not None:
                self.city = city.text

            region = address.find('StateOrProvince')
            if region is not None:
                self.region = region.text

            postcode = address.find('PostCode')
            if postcode is not None:
                self.postcode = postcode.text

            country = address.find('Country')
            if country is not None:
                self.country = country.text

        organization = elem.find('ContactPersonPrimary/ContactOrganization')
        if organization is not None:
            self.organization = organization.text
        else:
            self.organization = None

        position = elem.find('ContactPosition')
        if position is not None:
            self.position = position.text
        else:
            self.position = None



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2010 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

""" FGDC metadata parser """

from owslib.etree import etree
from owslib import util


class Metadata(object):
    """ Process metadata """
    def __init__(self, md):
        if hasattr(md, 'getroot'):  # standalone document
            self.xml = etree.tostring(md.getroot())
        else:  # part of a larger document
            self.xml = etree.tostring(md)

        self.idinfo = Idinfo(md)
        self.eainfo = Eainfo(md)
        self.distinfo = Distinfo(md)
        self.metainfo = Metainfo(md)

        if self.idinfo.datasetid:
            self.identifier = self.idinfo.datasetid


class Idinfo(object):
    """ Process idinfo """
    def __init__(self, md):
        val = md.find('idinfo/datasetid')
        self.datasetid = util.testXMLValue(val)

        val = md.find('idinfo/citation')
        self.citation = Citation(val)

        val = md.find('idinfo/descript')
        if val is not None:
            self.descript = Descript(val)

        val = md.find('idinfo/timeperd')
        self.timeperd = Timeperd(val)

        val = md.find('idinfo/status')
        if val is not None:
            self.status = Status(val)

        val = md.find('idinfo/spdom')
        if val is not None:
            self.spdom = Spdom(val)

        val = md.find('idinfo/keywords')
        if val is not None:
            self.keywords = Keywords(val)

        val = md.find('idinfo/accconst')
        self.accconst = util.testXMLValue(val)

        val = md.find('idinfo/useconst')
        self.useconst = util.testXMLValue(val)

        val = md.find('idinfo/ptcontac')
        if val is not None:
            self.ptcontac = Ptcontac(val)

        val = md.find('idinfo/datacred')
        self.datacred = util.testXMLValue(val)

        val = md.find('idinfo/crossref')
        self.crossref = Citation(val)


class Citation(object):
    """ Process citation """
    def __init__(self, md):
        if md is not None:
            self.citeinfo = {}

            val = md.find('citeinfo/origin')
            self.citeinfo['origin'] = util.testXMLValue(val)

            val = md.find('citeinfo/pubdate')
            self.citeinfo['pubdate'] = util.testXMLValue(val)

            val = md.find('citeinfo/title')
            self.citeinfo['title'] = util.testXMLValue(val)

            val = md.find('citeinfo/geoform')
            self.citeinfo['geoform'] = util.testXMLValue(val)

            val = md.find('citeinfo/pubinfo/pubplace')
            self.citeinfo['pubplace'] = util.testXMLValue(val)

            val = md.find('citeinfo/pubinfo/publish')
            self.citeinfo['publish'] = util.testXMLValue(val)

            self.citeinfo['onlink'] = []
            for link in md.findall('citeinfo/onlink'):
                self.citeinfo['onlink'].append(util.testXMLValue(link))


class Descript(object):
    """ Process descript """
    def __init__(self, md):
        val = md.find('abstract')
        self.abstract = util.testXMLValue(val)

        val = md.find('purpose')
        self.purpose = util.testXMLValue(val)

        val = md.find('supplinf')
        self.supplinf = util.testXMLValue(val)


class Timeperd(object):
    """ Process timeperd """
    def __init__(self, md):
        if md is not None:
            val = md.find('current')
            self.current = util.testXMLValue(val)

            val = md.find('timeinfo')
            if val is not None:
                self.timeinfo = Timeinfo(val)


class Timeinfo(object):
    """ Process timeinfo """
    def __init__(self, md):
        val = md.find('sngdate')
        if val is not None:
            self.sngdate = Sngdate(val)

        val = md.find('rngdates')
        if val is not None:
            self.rngdates = Rngdates(val)


class Sngdate(object):
    """ Process sngdate """
    def __init__(self, md):
        val = md.find('caldate')
        self.caldate = util.testXMLValue(val)
        val = md.find('time')
        self.time = util.testXMLValue(val)


class Rngdates(object):
    """ Process rngdates """
    def __init__(self, md):
        val = md.find('begdate')
        self.begdate = util.testXMLValue(val)
        val = md.find('begtime')
        self.begtime = util.testXMLValue(val)
        val = md.find('enddate')
        self.enddate = util.testXMLValue(val)
        val = md.find('endtime')
        self.endtime = util.testXMLValue(val)


class Status(object):
    """ Process status """
    def __init__(self, md):
        val = md.find('progress')
        self.progress = util.testXMLValue(val)

        val = md.find('update')
        self.update = util.testXMLValue(val)


class Spdom(object):
    """ Process spdom """
    def __init__(self, md):
        val = md.find('bounding/westbc')
        self.westbc = util.testXMLValue(val)

        val = md.find('bounding/eastbc')
        self.eastbc = util.testXMLValue(val)

        val = md.find('bounding/northbc')
        self.northbc = util.testXMLValue(val)

        val = md.find('bounding/southbc')
        self.southbc = util.testXMLValue(val)

        if self.southbc is not None and \
                self.northbc is not None and \
                self.eastbc is not None and \
                self.westbc is not None:
            self.bbox = Bbox(self)


class Bbox(object):
    """ Generate bbox for spdom (convenience function) """
    def __init__(self, spdom):
        self.minx = spdom.westbc
        self.miny = spdom.southbc
        self.maxx = spdom.eastbc
        self.maxy = spdom.northbc


class Keywords(object):
    """ Process keywords """
    def __init__(self, md):
        self.theme = []
        self.place = []
        self.temporal = []

        for i in md.findall('theme'):
            theme = {}
            val = i.find('themekt')
            theme['themekt'] = util.testXMLValue(val)
            theme['themekey'] = []
            for j in i.findall('themekey'):
                themekey = util.testXMLValue(j)
                if themekey is not None:
                    theme['themekey'].append(themekey)
            self.theme.append(theme)

        for i in md.findall('place'):
            theme = {}
            place = {}
            val = i.find('placekt')
            theme['placekt'] = util.testXMLValue(val)
            theme['placekey'] = []
            for j in i.findall('placekey'):
                theme['placekey'].append(util.testXMLValue(j))
            self.place.append(place)

        for i in md.findall('temporal'):
            theme = {}
            temporal = {}
            val = i.find('tempkt')
            theme['tempkt'] = util.testXMLValue(val)
            theme['tempkey'] = []
            for j in i.findall('tempkey'):
                theme['tempkey'].append(util.testXMLValue(j))
            self.temporal.append(temporal)


class Ptcontac(object):
    """ Process ptcontac """
    def __init__(self, md):
        val = md.find('cntinfo/cntorgp/cntorg')
        self.cntorg = util.testXMLValue(val)

        val = md.find('cntinfo/cntorgp/cntper')
        self.cntper = util.testXMLValue(val)

        val = md.find('cntinfo/cntpos')
        self.cntpos = util.testXMLValue(val)

        val = md.find('cntinfo/cntaddr/addrtype')
        self.addrtype = util.testXMLValue(val)

        val = md.find('cntinfo/cntaddr/address')
        self.address = util.testXMLValue(val)

        val = md.find('cntinfo/cntaddr/city')
        self.city = util.testXMLValue(val)

        val = md.find('cntinfo/cntaddr/state')
        self.state = util.testXMLValue(val)

        val = md.find('cntinfo/cntaddr/postal')
        self.postal = util.testXMLValue(val)

        val = md.find('cntinfo/cntaddr/country')
        self.country = util.testXMLValue(val)

        val = md.find('cntinfo/cntvoice')
        self.voice = util.testXMLValue(val)

        val = md.find('cntinfo/cntemail')
        self.email = util.testXMLValue(val)


class Eainfo(object):
    """ Process eainfo """
    def __init__(self, md):
        val = md.find('eainfo/detailed/enttyp/enttypl')
        self.enttypl = util.testXMLValue(val)

        val = md.find('eainfo/detailed/enttyp/enttypd')
        self.enttypd = util.testXMLValue(val)

        val = md.find('eainfo/detailed/enttyp/enttypds')
        self.enttypds = util.testXMLValue(val)

        self.attr = []
        for i in md.findall('eainfo/detailed/attr'):
            attr = {}
            val = i.find('attrlabl')
            attr['attrlabl'] = util.testXMLValue(val)

            val = i.find('attrdef')
            attr['attrdef'] = util.testXMLValue(val)

            val = i.find('attrdefs')
            attr['attrdefs'] = util.testXMLValue(val)

            val = i.find('attrdomv/udom')
            attr['udom'] = util.testXMLValue(val)

            self.attr.append(attr)


class Distinfo(object):
    """ Process distinfo """
    def __init__(self, md):
        val = md.find('distinfo')
        if val is not None:
            val2 = val.find('stdorder')
            if val2 is not None:
                self.stdorder = {'digform': []}
                for link in val2.findall('digform'):
                    digform = {}
                    digform['name'] = util.testXMLValue(link.find('digtinfo/formname'))
                    digform['url'] = util.testXMLValue(link.find('digtopt/onlinopt/computer/networka/networkr/'))
                    self.stdorder['digform'].append(digform)


class Metainfo(object):
    """ Process metainfo """
    def __init__(self, md):
        val = md.find('metainfo/metd')
        self.metd = util.testXMLValue(val)

        val = md.find('metainfo/metrd')
        self.metrd = util.testXMLValue(val)

        val = md.find('metainfo/metc')
        if val is not None:
            self.metc = Ptcontac(val)

        val = md.find('metainfo/metstdn')
        self.metstdn = util.testXMLValue(val)

        val = md.find('metainfo/metstdv')
        self.metstdv = util.testXMLValue(val)

        val = md.find('metainfo/metac')
        self.metac = util.testXMLValue(val)

        val = md.find('metainfo/metuc')
        self.metuc = util.testXMLValue(val)
